# Neurodamus
# Copyright 2018 - Blue Brain Project, EPFL

from __future__ import absolute_import
import gc
import glob
import itertools
import logging
import math
import os
import subprocess
from os import path as ospath
from collections import namedtuple, defaultdict
from contextlib import contextmanager
from shutil import copyfileobj, move

from .core import MPI, mpi_no_errors, return_neuron_timings, run_only_rank0
from .core import NeurodamusCore as Nd
from .core.configuration import CircuitConfig, Feature, GlobalConfig, SimConfig
from .core._engine import EngineBase
from .core._shmutils import SHMUtil
from .core.configuration import ConfigurationError, find_input_file, get_debug_cell_gid
from .core.nodeset import PopulationNodes
from .cell_distributor import CellDistributor, VirtualCellPopulation, GlobalCellManager
from .cell_distributor import LoadBalance, LoadBalanceMode
from .connection_manager import SynapseRuleManager, edge_node_pop_names
from .gap_junction import GapJunctionManager
from .replay import MissingSpikesPopulationError, SpikeManager
from .stimulus_manager import StimulusManager
from .modification_manager import ModificationManager
from .neuromodulation_manager import NeuroModulationManager
from .target_manager import TargetSpec, TargetManager
from .utils import compat
from .utils.logging import log_stage, log_verbose, log_all
from .utils.memory import DryRunStats, trim_memory, pool_shrink, free_event_queues, print_mem_usage
from .utils.timeit import TimerManager, timeit
from .core.coreneuron_configuration import CoreConfig
# Internal Plugins
from . import ngv as _ngv  # NOQA


class METypeEngine(EngineBase):
    CellManagerCls = CellDistributor
    InnerConnectivityCls = SynapseRuleManager
    ConnectionTypes = {
        None: SynapseRuleManager,
        "Synaptic": SynapseRuleManager,
        "GapJunction": GapJunctionManager,
        "NeuroModulation": NeuroModulationManager
    }
    CircuitPrecedence = 0


class CircuitManager:
    """
    Holds and manages populations and associated nodes and edges

    For backward compat, base population doesnt have a population name (it is '')
    All other nodes must have a name, read from sonata pop name, or the BlueConfig circuit
    As so, Sonata is preferred when using multiple node files
    """

    def __init__(self):
        self.node_managers = {}  # dict {pop_name -> cell_manager}  # nrn pop is None
        self.virtual_node_managers = {}   # same, but for virtual ones (no cells)
        self.edge_managers = defaultdict(list)  # dict {(src_pop, dst_pop) -> list[synapse_manager]}
        self.alias = {}          # dict {name -> pop_name}
        self.global_manager = GlobalCellManager()
        self.global_target = TargetManager.create_global_target()

    def initialized(self):
        return bool(self.node_managers)

    def register_node_manager(self, cell_manager):
        pop = cell_manager.population_name
        if pop in self.node_managers:
            raise ConfigurationError("Already existing node manager for population %s" % pop)
        self.node_managers[pop] = cell_manager
        self.alias[cell_manager.circuit_name] = pop
        self.global_manager.register_manager(cell_manager)
        if cell_manager.is_initialized():
            self.global_target.append_nodeset(cell_manager.local_nodes)

    def _new_virtual_node_manager(self, circuit):
        """Instantiate a new virtual node manager explicitly."""
        # Only happens with Sonata config files
        import libsonata
        storage = libsonata.NodeStorage(circuit.CellLibraryFile)
        pop_name, _ = circuit.CircuitTarget.split(":")  # Sonata config fills population
        node_size = storage.open_population(pop_name).size
        gid_vec = list(range(1, node_size+1))
        virtual_cell_manager = VirtualCellPopulation(pop_name, gid_vec)
        self.virtual_node_managers[pop_name] = virtual_cell_manager
        self.global_target.append_nodeset(virtual_cell_manager.local_nodes)
        return virtual_cell_manager

    @staticmethod
    def new_node_manager_bare(circuit: CircuitConfig, target_manager, run_conf, **kwargs):
        engine = circuit.Engine or METypeEngine
        CellManagerCls = engine.CellManagerCls or CellDistributor
        return CellManagerCls(circuit, target_manager, run_conf, **kwargs)

    def new_node_manager(self, circuit, target_manager, run_conf, *, load_balancer=None, **kwargs):
        if circuit.get("PopulationType") == "virtual":
            return self._new_virtual_node_manager(circuit)
        cell_manager = self.new_node_manager_bare(circuit, target_manager, run_conf, **kwargs)
        cell_manager.load_nodes(load_balancer, **kwargs)
        self.register_node_manager(cell_manager)
        return cell_manager

    def get_node_manager(self, name):
        name = self.alias.get(name, name)
        return self.node_managers.get(name)

    def has_population(self, pop_name):
        return pop_name in self.node_managers

    def unalias_pop_keys(self, source, destination):
        """Un-alias population names"""
        return self.alias.get(source, source), self.alias.get(destination, destination)

    def get_edge_managers(self, source, destination):
        edge_pop_keys = self.unalias_pop_keys(source, destination)
        return self.edge_managers.get(edge_pop_keys) or []

    def get_edge_manager(self, source, destination, conn_type=SynapseRuleManager):
        managers = [manager for manager in self.get_edge_managers(source, destination)
                    if isinstance(manager, conn_type)]
        return managers[0] if managers else None

    def get_create_edge_manager(self, conn_type, source, destination, src_target,
                                manager_args=(), **kw):
        source, destination = self.unalias_pop_keys(source, destination)
        manager = self.get_edge_manager(source, destination, conn_type)
        if manager:
            return manager

        if not self.has_population(destination):
            # This is likely an error, except...
            if src_target.population == '' and self.has_population(''):
                logging.warning("Sonata Edges target population %s was not found. "
                                "Since base population is unknown, assuming that's the target.\n"
                                "To silence this warning please switch to Sonata nodes or specify "
                                "the base population by prefixing CircuitTarget with pop_name:",
                                destination)
                self.alias[destination] = ''
                destination = ''
                source = self.alias.get(source)  # refresh unaliasing
            else:
                raise ConfigurationError("Can't find projection Node population: " + destination)

        src_manager = self.node_managers.get(source) or self.virtual_node_managers.get(source)
        if src_manager is None:  # src manager may not exist -> virtual
            log_verbose("No known population %s. Creating Virtual src for projection", source)
            if conn_type not in (SynapseRuleManager, _ngv.GlioVascularManager):
                raise ConfigurationError("Custom connections require instantiated source nodes")
            src_manager = VirtualCellPopulation(source, None, src_target.name)

        target_cell_manager = kw["cell_manager"] = self.node_managers[destination]
        kw["src_cell_manager"] = src_manager
        manager = conn_type(*manager_args, **kw)
        self.edge_managers[(source, destination)].append(manager)
        target_cell_manager.register_connection_manager(manager)
        return manager

    def all_node_managers(self):
        return self.node_managers.values()

    def all_synapse_managers(self):
        return itertools.chain.from_iterable(self.edge_managers.values())

    @property
    def base_cell_manager(self):
        return self.get_node_manager(None)

    @run_only_rank0
    def write_population_offsets(self):
        """Write infor in <output_dir>/populations_offset.dat
        format population name::gid offset::population alias
        The virtual population offset is also written for synapse replay in restore.
        """
        with open(self._pop_offset_file(create=True), "w") as f:
            pop_offsets, alias_pop = self.get_population_offsets()
            virtual_pop_offsets = self.get_virtual_population_offsets()
            for alias, pop in alias_pop.items():
                f.write("{}::{}::{}\n".format(pop or ' ', pop_offsets[pop], alias or ' '))
            for pop, offset in virtual_pop_offsets.items():
                f.write("{}::{}::{}\n".format(pop, offset, 'virtual'))

    def get_population_offsets(self):
        pop_offsets = {pop_name: node_manager.local_nodes.offset
                       for pop_name, node_manager in self.node_managers.items()}
        alias_pop = {alias: pop_name for alias, pop_name in self.alias.items()}
        return pop_offsets, alias_pop

    def get_virtual_population_offsets(self):
        pop_offsets = {pop_name: node_manager.local_nodes.offset
                       for pop_name, node_manager in self.virtual_node_managers.items()}
        return pop_offsets

    @classmethod
    def read_population_offsets(cls, read_virtual_pop=False):
        """Read population offsets from populations_offset.dat
        Args:
            read_virtual_pop: read virtual population info (for replay), no need for reporting
        """
        pop_offsets = {}
        alias_pop = {}
        pop_offset_file = cls._pop_offset_file()
        if SimConfig.restore_coreneuron:
            if MPI.rank == 0 and not os.path.exists(pop_offset_file):
                # RESTORE: link to populations_offset.dat besides save directory
                os.symlink(os.path.join(SimConfig.restore, "../populations_offset.dat"),
                           pop_offset_file)
            MPI.barrier()
        with open(pop_offset_file, "r") as f:
            offsets = [line.strip().split("::") for line in f]
            for entry in offsets:
                if not read_virtual_pop and entry[2] == 'virtual':
                    continue
                pop_offsets[entry[0] or None] = int(entry[1])
                alias_pop[entry[2] or None] = entry[0] or None
        return pop_offsets, alias_pop

    @classmethod
    def _pop_offset_file(self, create=False):
        outdir = ospath.join(SimConfig.output_root)
        create and os.makedirs(outdir, exist_ok=True)
        return ospath.join(outdir, "populations_offset.dat")

    def __del__(self):
        """ De-init. Edge managers must be destructed first """
        del self.edge_managers
        del self.virtual_node_managers
        del self.node_managers


class Node:
    """The Node class is the main entity for a distributed neurodamus execution.

    It internally instantiates parallel structures and distributes the cells among all the nodes.
    It is relatively low-level, for a standard run consider using the Neurodamus class instead.
    """

    _default_population = 'All'
    """The default population name for e.g. Reports."""

    def __init__(self, config_file, options=None):
        """ Creates a neurodamus executor
        Args:
            config_file: A Sonata config file
            options: A dictionary of run options typically coming from cmd line
        """
        if config_file:
            if config_file.endswith("BlueConfig"):
                raise ConfigurationError(
                    "Legacy format BlueConfig is not supported, please migrate to SONATA config")
            import libsonata
            conf = libsonata.SimulationConfig.from_file(config_file)
            Nd.init(log_filename=conf.output.log_file)
        else:
            Nd.init()  # ensure/load neurodamus mods
        self._run_conf: dict  # Multi-cycle runs preserve this

        # The Recipe being None is allowed internally for e.g. setting up multi-cycle runs
        # It shall not be used as Public API
        if config_file is not None:
            # This is global initialization, happening once, regardless of number of cycles
            log_stage("Setting up Neurodamus configuration")
            self._pc = Nd.pc
            self._spike_vecs = []
            self._spike_populations = []
            Nd.execute("cvode = new CVode()")
            SimConfig.init(config_file, options)
            if SimConfig.use_coreneuron:
                CoreConfig.output_root = SimConfig.output_root
                CoreConfig.datadir = SimConfig.coreneuron_datadir
                # Instantiate the CoreNEURON artificial cell object which is used to fill up
                # the empty ranks. This need to be done before the circuit is finitialized
                CoreConfig.instantiate_artificial_cell()
            self._run_conf = SimConfig.run_conf
            self._target_manager = TargetManager(self._run_conf)
            self._target_spec = TargetSpec(self._run_conf.get("CircuitTarget"))
            if SimConfig.use_neuron:
                self._sonatareport_helper = Nd.SonataReportHelper(Nd.dt, True)
            self._base_circuit: CircuitConfig = SimConfig.base_circuit
            self._extra_circuits = SimConfig.extra_circuits
            self._pr_cell_gid = get_debug_cell_gid(options)
            self._core_replay_file = ""
            self._is_ngv_run = any(c.Engine.__name__ == "NGVEngine"
                                   for c in self._extra_circuits.values() if c.Engine)
            self._initial_rss = 0
            self._cycle_i = 0
            self._n_cycles = 1
            self._shm_enabled = False
            self._dry_run_stats = None
        else:
            self._run_conf  # Assert this is defined (if not multicyle runs are not properly set)

        # Init unconditionally
        self._circuits = CircuitManager()
        self._stim_list = None
        self._report_list = None
        self._stim_manager = None
        self._elec_manager = None
        self._sim_ready = False
        self._jumpstarters = []
        self._cell_state_dump_t = None
        self._bbss = Nd.BBSaveState()

        # Register the global target and cell manager
        self._target_manager.register_target(self._circuits.global_target)
        self._target_manager.init_hoc_manager(self._circuits.global_manager)

    #
    # public 'read-only' properties - object modification on user responsibility
    circuits = property(lambda self: self._circuits)
    target_manager = property(lambda self: self._target_manager)
    stim_manager = property(lambda self: self._stim_manager)
    elec_manager = property(lambda self: self._elec_manager)
    stims = property(lambda self: self._stim_list)
    reports = property(lambda self: self._report_list)

    def all_circuits(self, exclude_disabled=True):
        if not exclude_disabled or self._base_circuit.CircuitPath:
            yield self._base_circuit
        yield from self._extra_circuits.values()

    # -
    def load_targets(self):
        """Initialize targets. Nodesets are loaded on demand.
        """
        # If a base population is specified register it before targets to create on demand
        base_population = self._run_conf.get("BasePopulation")
        if base_population:
            logging.info("Default population selected: %s", base_population)
            PopulationNodes.create_pop(base_population, is_base_pop=True)

        for circuit in self.all_circuits():
            log_verbose("Loading targets for circuit %s", circuit._name or "(default)")
            self._target_manager.load_targets(circuit)

        self._target_manager.load_user_target()

    # -
    @mpi_no_errors
    @timeit(name="Compute LB")
    def compute_load_balance(self):
        """In case the user requested load-balance this function instantiates a
        CellDistributor to split cells and balance those pieces across the available CPUs.
        """
        log_stage("Computing Load Balance")
        is_sonata_config = SimConfig.is_sonata_config
        circuit = self._base_circuit
        if is_sonata_config:
            for name, circuit in self._extra_circuits.items():
                if circuit.get("PopulationType") != "virtual":
                    break
            if circuit.get("PopulationType") == "virtual":
                logging.warning(
                    "Cannot calculate the load balance because only virtual populations were found"
                )
                return None
            logging.info("Activating experimental LB for Sonata circuit '%s'", name)

        if not circuit.CircuitPath:
            logging.info(" => No circuit for Load Balancing. Skipping... ")
            return None

        _ = PopulationNodes.offset_freezer()  # Dont offset while in loadbal

        # Info about the cells to be distributed
        target_spec = TargetSpec(circuit.CircuitTarget)
        target = self.target_manager.get_target(target_spec)

        # Check / set load balance mode
        lb_mode = LoadBalance.select_lb_mode(SimConfig, self._run_conf, target)
        if lb_mode == LoadBalanceMode.RoundRobin:
            return None

        # Build load balancer as per requested options
        # Compat Note:
        # data_src in BlueConfig mode was the nrnPath. Not anymore (except if not defined or <NONE>)
        prosp_hosts = self._run_conf.get("ProspectiveHosts")
        data_src = (
            circuit.CircuitPath if is_sonata_config
            else self._run_conf["nrnPath"] or circuit.CircuitPath
        )
        pop = target_spec.population
        load_balancer = LoadBalance(lb_mode, data_src, pop, self._target_manager, prosp_hosts)

        if load_balancer.valid_load_distribution(target_spec):
            logging.info("Load Balancing done.")
            return load_balancer

        logging.info("Could not reuse load balance data. Doing a Full Load-Balance")
        cell_dist = self._circuits.new_node_manager(circuit, self._target_manager, self._run_conf)
        with load_balancer.generate_load_balance(target_spec, cell_dist):
            # Instantiate a basic circuit to evaluate complexities
            cell_dist.finalize()
            self._circuits.global_manager.finalize()
            SimConfig.update_connection_blocks(self._circuits.alias)
            target_manager = self._target_manager
            self._create_synapse_manager(SynapseRuleManager, self._base_circuit, target_manager)

        # reset since we instantiated with RR distribution
        Nd.t = .0  # Reset time
        self.clear_model()

        return load_balancer

    # -
    @mpi_no_errors
    @timeit(name="Cell creation")
    def create_cells(self, load_balance=None):
        """Instantiate and distributes the cells of the network.
        Any targets will be updated to know which cells are local to the cpu.
        """
        # We wont go further if ProspectiveHosts is defined to some other cpu count
        prosp_hosts = self._run_conf.get("ProspectiveHosts")
        if load_balance and prosp_hosts not in (None, MPI.size):
            logging.warning(
                "Load Balance requested for %d CPUs (as per ProspectiveHosts). "
                "To continue execution launch on a partition of that size",
                prosp_hosts)
            Nd.quit(1)

        if SimConfig.dry_run:
            logging.info("Memory usage after inizialization:")
            print_mem_usage()
            self._dry_run_stats = DryRunStats()
            self._dry_run_stats.try_import_cell_memory_usage()
            loader_opts = {"dry_run_stats": self._dry_run_stats}
        else:
            loader_opts = {}

        # Check dynamic attributes required before loading cells
        SimConfig.check_cell_requirements(self.target_manager)

        log_stage("LOADING NODES")
        config = SimConfig.cli_options
        if not load_balance:
            logging.info("Load-balance object not present. Continuing Round-Robin...")

        # Always create a cell_distributor even if engine is disabled.
        # Fake CoreNeuron cells are created in it
        cell_distributor = CellDistributor(self._base_circuit, self._target_manager, self._run_conf)
        cell_distributor.load_nodes(load_balance, loader_opts=loader_opts)  # no-op if disabled
        self._circuits.register_node_manager(cell_distributor)

        # SUPPORT for extra/custom Circuits
        for name, circuit in self._extra_circuits.items():
            log_stage("Circuit %s", name)
            if config.restrict_node_populations and name not in config.restrict_node_populations:
                logging.warning("Skipped node population (restrict_node_populations)")
                continue
            self._circuits.new_node_manager(circuit, self._target_manager, self._run_conf,
                                            load_balancer=load_balance,
                                            loader_opts=loader_opts)

        lfp_weights_file = self._run_conf.get("LFPWeightsPath")
        if lfp_weights_file:
            if SimConfig.use_coreneuron:
                lfp_manager = self._circuits.global_manager._lfp_manager
                cell_managers = self._circuits.global_manager._cell_managers
                population_list = [
                    manager.population_name
                    for manager in cell_managers
                    if manager.population_name is not None
                ]
                lfp_manager.load_lfp_config(lfp_weights_file, population_list)
            else:
                logging.warning("Online LFP supported only with CoreNEURON.")

        PopulationNodes.freeze_offsets()  # Dont offset further, could change gids

        # Let the cell managers have any final say in the cell objects
        log_stage("FINALIZING CIRCUIT CELLS")

        for cell_manager in self._circuits.all_node_managers():
            log_stage("Circuit %s", cell_manager.circuit_name or "(default)")
            if SimConfig.dry_run:
                cell_manager.finalize(dry_run_stats_obj=self._dry_run_stats)
            else:
                cell_manager.finalize()

        if SimConfig.dry_run:
            self._dry_run_stats.collect_all_mpi()
            self._dry_run_stats.export_cell_memory_usage()
            self._dry_run_stats.estimate_cell_memory()

        # Final bits after we have all cell managers
        self._circuits.global_manager.finalize()
        SimConfig.update_connection_blocks(self._circuits.alias)

    # -
    @mpi_no_errors
    @timeit(name="Synapse creation")
    def create_synapses(self):
        """Create synapses among the cells, handling connections that appear in the config file
        """
        log_stage("LOADING CIRCUIT CONNECTIVITY")
        target_manager = self._target_manager
        manager_kwa = {"load_offsets": self._is_ngv_run, "dry_run_stats": self._dry_run_stats}

        if circuit := self._base_circuit:
            self._create_synapse_manager(SynapseRuleManager, circuit, target_manager, **manager_kwa)

        for circuit in self._extra_circuits.values():
            Engine = circuit.Engine or METypeEngine
            SynManagerCls = Engine.InnerConnectivityCls
            self._create_synapse_manager(SynManagerCls, circuit, target_manager, **manager_kwa)

        MPI.check_no_errors()
        log_stage("Handling projections...")
        for pname, projection in SimConfig.projections.items():
            self._load_projections(pname, projection, **manager_kwa)

        if SimConfig.dry_run:
            self.syn_total_memory = self._dry_run_stats.collect_display_syn_counts()
            return

        log_stage("Configuring connections...")
        for conn_conf in SimConfig.connections.values():
            self._process_connection_configure(conn_conf)

        logging.info("Done, but waiting for all ranks")

    def _create_synapse_manager(self, ctype, conf, *args, **kwargs):
        """Create a synapse manager for intra-circuit connectivity
        """
        log_stage("Circuit %s", conf._name or "(default)")
        if not conf.get("nrnPath"):
            logging.info(" => No connectivity set as internal. See projections")
            return

        if SimConfig.cli_options.restrict_connectivity >= 2:
            logging.warning("Skipped connectivity (restrict_connectivity)")
            return

        c_target = TargetSpec(conf.get("CircuitTarget"))
        if c_target.population is None:
            c_target.population = self._circuits.alias.get(conf._name)

        edge_file, *pop = conf.get("nrnPath").split(":")
        edge_pop = pop[0] if pop else None
        if not os.path.isabs(edge_file):
            edge_file = find_input_file(edge_file)
        src, dst = edge_node_pop_names(edge_file, edge_pop)

        logging.info("Processing edge file %s, pop: %s", edge_file, edge_pop)

        if src and dst and src != dst:
            raise ConfigurationError("Inner connectivity with different populations")

        dst = self.circuits.alias.get(dst, dst)
        if dst not in SimConfig.cli_options.restrict_node_populations:
            logging.warning("Skipped connectivity (restrict_node_populations)")
            return

        manager = self._circuits.get_create_edge_manager(
            ctype, src, dst, c_target, (conf, *args), **kwargs
        )
        self._load_connections(conf, manager)  # load internal connections right away

    def _process_connection_configure(self, conn_conf):
        source_t = TargetSpec(conn_conf["Source"])
        dest_t = TargetSpec(conn_conf["Destination"])
        source_t.population, dest_t.population = self._circuits.unalias_pop_keys(
            source_t.population, dest_t.population
        )
        src_target = self.target_manager.get_target(source_t)
        dst_target = self.target_manager.get_target(dest_t)
        # Loop over population pairs
        for src_pop in src_target.population_names:
            for dst_pop in dst_target.population_names:
                # Loop over all managers having connections between the populations
                for conn_manager in self._circuits.get_edge_managers(src_pop, dst_pop):
                    logging.debug("Using connection manager: %s", conn_manager)
                    conn_manager.configure_connections(conn_conf)

    # -
    def _load_connections(self, circuit_conf, conn_manager):
        if conn_manager.is_file_open:  # Base connectivity
            conn_manager.create_connections()

        # Continue support for compatibility, but new BlueConfigs should use Projection blocks
        bonus_file = circuit_conf.get("BonusSynapseFile")
        if bonus_file:
            logging.info(" > Loading Circuit BonusSynapseFile")
            n_synapse_files = int(circuit_conf.get("NumBonusFiles", 1))
            conn_manager.open_synapse_file(bonus_file, None, n_synapse_files)
            conn_manager.create_connections()

    # -
    @mpi_no_errors
    def _load_projections(self, pname, projection, **kw):
        """Check for Projection blocks
        """
        target_manager = self._target_manager
        projection = compat.Map(projection).as_dict(True)
        ptype = projection.get("Type")  # None, GapJunctions, NeuroGlial, NeuroModulation...
        ptype_cls = EngineBase.connection_types.get(ptype)
        source_t = TargetSpec(projection.get("Source"))
        dest_t = TargetSpec(projection.get("Destination"))

        if SimConfig.cli_options.restrict_connectivity >= 1:
            logging.warning("Skipped projections %s->%s (restrict_connectivity)", source_t, dest_t)
            return

        if not ptype_cls:
            raise RuntimeError("No Engine to handle connectivity of type '%s'" % ptype)

        ppath, *pop_name = projection["Path"].split(":")
        edge_pop_name = pop_name[0] if pop_name else None
        if not os.path.exists(ppath):
            ppath = self._find_projection_file(ppath)

        logging.info("Processing Edge file: %s", ppath)

        # Update the target spec with the actual populations
        src_pop, dst_pop = edge_node_pop_names(
            ppath, edge_pop_name, source_t.population, dest_t.population
        )
        source_t.population, dest_t.population = self._circuits.unalias_pop_keys(src_pop, dst_pop)
        src_target = self.target_manager.get_target(source_t)
        dst_target = self.target_manager.get_target(dest_t)

        # If the src_pop is not a known node population, allow creating a Virtual one
        src_populations = src_target.population_names or [source_t.population]

        for src_pop in src_populations:
            for dst_pop in dst_target.population_names:
                logging.info(" * %s (Type: %s, Src: %s, Dst: %s)", pname, ptype, src_pop, dst_pop)
                conn_manager = self._circuits.get_create_edge_manager(
                    ptype_cls, src_pop, dst_pop, source_t,
                    (projection, target_manager), **kw  # args to ptype_cls if creating
                )
                logging.debug("Using connection manager: %s", conn_manager)
                proj_source = ":".join([ppath] + pop_name)
                conn_manager.open_edge_location(proj_source, projection, src_name=src_pop)
                conn_manager.create_connections(source_t.name, dest_t.name)

    # -
    def _find_projection_file(self, proj_path):
        """Determine the full path to a projection.
        The "Path" might specify the filename. If not, it will attempt the old 'proj_nrn.h5'
        """
        return self._find_config_file(
            proj_path, ("ProjectionPath", "CircuitPath"), alt_filename="proj_nrn.h5")

    def _find_config_file(self, filepath, path_conf_entries=(), alt_filename=None):
        search_paths = [self._run_conf[path_key]
                        for path_key in path_conf_entries
                        if self._run_conf.get(path_key)]
        return find_input_file(filepath, search_paths, alt_filename)

    # -
    @mpi_no_errors
    @timeit(name="Enable Stimulus")
    def enable_stimulus(self):
        """Iterate over any stimuli/stim injects defined in the config file given to the simulation
        and instantiate them.
        This iterates over the injects, getting the stim/target combinations
        and passes the raw text in field/value pairs to a StimulusManager object to interpret the
        text and instantiate an actual stimulus object.
        """
        if Feature.Stimulus not in SimConfig.cli_options.restrict_features:
            logging.warning("Skipped Stimulus (restrict_features)")
            return

        log_stage("Stimulus Apply.")

        # Setup of Electrode objects part of enable stimulus
        self._enable_electrodes()

        # for each stimulus defined in the config file, request the stimmanager to instantiate
        self._stim_manager = StimulusManager(self._target_manager, self._elec_manager)

        # build a dictionary of stims for faster lookup : useful when applying 10k+ stims
        # while we are at it, check if any stims are using extracellular
        has_extra_cellular = False
        stim_dict = {}
        for stim_name, stim in SimConfig.stimuli.items():
            stim = compat.Map(stim)
            if stim_name in stim_dict:
                raise ConfigurationError("Stimulus declared more than once: %s", stim_name)
            stim_dict[stim_name] = stim
            if stim.get("Mode") == "Extracellular":
                has_extra_cellular = True

        # Treat extracellular stimuli
        if has_extra_cellular:
            self._stim_manager.interpret_extracellulars(SimConfig.injects, SimConfig.stimuli)

        logging.info("Instantiating Stimulus Injects:")

        for name, inject in SimConfig.injects.items():
            target_spec = TargetSpec(inject.get("Target")) if isinstance(inject.get("Target"), str)\
                else TargetSpec(inject.get("Target").s)
            stim_name = inject.get("Stimulus") if isinstance(inject.get("Stimulus"), str) \
                else inject.get("Stimulus").s
            stim = stim_dict.get(stim_name)
            if stim is None:
                raise ConfigurationError("Stimulus Inject %s uses non-existing Stim %s",
                                         name, stim_name)

            stim_pattern = stim["Pattern"]
            if stim_pattern == "SynapseReplay":
                continue  # Handled by enable_replay

            logging.info(" * [STIM] %s: %s (%s) -> %s", name, stim_name, stim_pattern, target_spec)
            self._stim_manager.interpret(target_spec, stim)

    # -
    def _enable_electrodes(self):
        if SimConfig.use_coreneuron:
            # Coreneuron doesnt support electrodes
            return False
        electrode_path = self._run_conf.get("ElectrodesPath")
        if electrode_path is not None:
            logging.info("ElectrodeManager using electrodes from %s", electrode_path)
        else:
            logging.info("No electrodes path. Extracellular class of stimuli will be unavailable")

        self._elec_manager = Nd.ElectrodeManager(
            electrode_path and Nd.String(electrode_path),
            SimConfig.get_blueconfig_hoc_section("parsedElectrodes")
        )

    # -
    @mpi_no_errors
    def enable_replay(self):
        """Activate replay according to BlueConfig. Call before connManager.finalize
        """
        if Feature.Replay not in SimConfig.cli_options.restrict_features:
            logging.warning("Skipped Replay (restrict_features)")
            return

        log_stage("Handling Replay")

        if SimConfig.use_coreneuron and bool(self._core_replay_file):
            logging.info(" -> [REPLAY] Reusing stim file from previous cycle")
            return

        replay_dict = {}
        for stim_name, stim in SimConfig.stimuli.items():
            pattern = stim.get("Pattern") if isinstance(stim.get("Pattern"), str) \
                else stim.get("Pattern").s
            if pattern == "SynapseReplay":
                replay_dict[stim_name] = compat.Map(stim).as_dict(parse_strings=True)

        for name, inject in SimConfig.injects.items():
            inject = compat.Map(inject).as_dict(parse_strings=True)
            target = inject["Target"]
            source = inject.get("Source")
            stim_name = inject["Stimulus"]
            connectivity_type = inject.get("Type")
            stim = replay_dict.get(stim_name)
            if stim is None:  # It's a non-replay inject. Injects are checked in enable_stimulus
                continue

            # Since saveUpdate merge there are two delay concepts:
            #  - shift: times are shifted (previous delay)
            #  - delay: Spike replays are suppressed until a certain time
            tshift = Nd.t if stim.get("Timing") == "Relative" else .0
            delay = stim.get("Delay", .0)
            logging.info(" * [SYN REPLAY] %s (%s -> %s, time shift: %d, delay: %d)",
                         name, stim_name, target, tshift, delay)
            self._enable_replay(source, target, stim, tshift, delay, connectivity_type)

    # -
    def _enable_replay(self, source, target, stim_conf, tshift=.0, delay=.0,
                       connectivity_type=None):
        spike_filepath = find_input_file(stim_conf["SpikeFile"])
        ptype_cls = EngineBase.connection_types.get(connectivity_type)
        src_target = self.target_manager.get_target(source)
        dst_target = self.target_manager.get_target(target)

        if SimConfig.restore_coreneuron:
            pop_offsets, alias_pop = CircuitManager.read_population_offsets(read_virtual_pop=True)

        for src_pop in src_target.population_names:
            try:
                log_verbose("Loading replay spikes for population '%s'", src_pop)
                spike_manager = SpikeManager(spike_filepath, tshift, src_pop)  # Disposable
            except MissingSpikesPopulationError:
                logging.info("  > No replay for src population: '%s'", src_pop)
                continue

            for dst_pop in dst_target.population_names:
                src_pop_str, dst_pop_str = src_pop or "(base)", dst_pop or "(base)"

                if SimConfig.restore_coreneuron:  # Node and Edges managers not initialized
                    src_pop_offset = pop_offsets[src_pop] if src_pop in pop_offsets \
                        else pop_offsets[alias_pop[src_pop]]
                else:
                    conn_manager = self._circuits.get_edge_manager(src_pop, dst_pop, ptype_cls)
                    if not conn_manager and SimConfig.cli_options.restrict_connectivity >= 1:
                        continue
                    assert conn_manager, f"Missing edge manager for {src_pop_str} -> {dst_pop_str}"
                    src_pop_offset = conn_manager.src_pop_offset

                logging.info("=> Population pathway %s -> %s. Source offset: %d",
                             src_pop_str, dst_pop_str, src_pop_offset)
                if SimConfig.use_coreneuron:
                    self._coreneuron_replay_append(spike_manager, src_pop_offset)
                else:
                    conn_manager.replay(spike_manager, source, target, delay)

    def _coreneuron_replay_append(self, spike_manager, gid_offset=None):
        """Write replay spikes in single file for CoreNeuron"""
        # To be loaded as PatternStim, requires final gids (with offset)
        # Initialize file if non-existing
        if not self._core_replay_file:
            self._core_replay_file = ospath.join(SimConfig.output_root, 'pattern.dat')
            if MPI.rank == 0:
                log_verbose("Creating pattern.dat file for CoreNEURON. Gid offset: %d", gid_offset)
                spike_manager.dump_ascii(self._core_replay_file, gid_offset)
        else:
            if MPI.rank == 0:
                log_verbose("Appending to pattern.dat. Gid offset: %d", gid_offset)
                with open(self._core_replay_file, "a") as f:
                    spike_manager.dump_ascii(f, gid_offset)

    # -
    @mpi_no_errors
    @timeit(name="Enable Modifications")
    def enable_modifications(self):
        """Iterate over any Modification blocks read from the BlueConfig and apply them to the
        network. The steps needed are more complex than NeuronConfigures, so the user should not be
        expected to write the hoc directly, but rather access a library of already available mods.
        """
        # mod_mananger gets destroyed when function returns (not required)
        # mod_manager = Nd.ModificationManager(self._target_manager.hoc)
        log_stage("Enabling modifications...")

        mod_manager = ModificationManager(self._target_manager)
        for name, mod in SimConfig.modifications.items():
            mod_info = compat.Map(mod)
            target_spec = TargetSpec(mod_info["Target"])
            logging.info(" * [MOD] %s: %s -> %s", name, mod_info["Type"], target_spec)
            mod_manager.interpret(target_spec, mod_info)

    # Reporting
    ReportParams = namedtuple("ReportParams", "name, rep_type, report_on, unit, format, dt, "
                              "start, end, output_dir, electrode, scaling, isc")

    # -
    # @mpi_no_errors - not required since theres a call inside before make_comm()
    @timeit(name="Enable Reports")
    def enable_reports(self):
        """Iterate over reports defined in BlueConfig and instantiate them.
        """
        log_stage("Reports Enabling")
        n_errors = 0
        reports_conf = SimConfig.reports
        self._report_list = []

        # Create a map of offsets so that it can be used even on coreneuron save-restore
        if self._circuits.initialized():
            self._circuits.write_population_offsets()
            pop_offsets_alias = self._circuits.get_population_offsets()
        else:
            pop_offsets_alias = CircuitManager.read_population_offsets()
        if SimConfig.use_coreneuron:
            CoreConfig.write_report_count(len(reports_conf))

        for rep_name, rep_conf in reports_conf.items():
            target_spec = TargetSpec(rep_conf["Target"])
            target = self._target_manager.get_target(target_spec)

            # Build final config. On errors log, stop only after all reports processed
            rep_params = self._report_build_params(rep_name, rep_conf, target, pop_offsets_alias)
            if rep_params is None:
                n_errors += 1
                continue

            if SimConfig.use_coreneuron and MPI.rank == 0:
                if not self._report_write_coreneuron_config(rep_name, rep_conf, target, rep_params):
                    n_errors += 1
                    continue

            if SimConfig.restore_coreneuron:
                continue  # we dont even need to initialize reports

            report = Nd.Report(*rep_params)

            if not SimConfig.use_coreneuron or rep_params.rep_type == "Synapse":
                self._report_setup(report, rep_conf, target, rep_params.rep_type)

            # Custom reporting. TODO: Move `_report_setup` to cellManager.enable_report
            target_population = target_spec.population or self._target_spec.population
            cell_manager = self._circuits.get_node_manager(target_population)
            cell_manager.enable_report(report, target, SimConfig.use_coreneuron)

            self._report_list.append(report)

        if n_errors > 0:
            raise ConfigurationError("%d reporting errors detected. Terminating" % (n_errors,))

        MPI.check_no_errors()

        self._reports_init(pop_offsets_alias)

        # electrode manager is no longer needed. free the memory
        if self._elec_manager is not None:
            self._elec_manager.clear()

    #
    def _report_build_params(self, rep_name, rep_conf, target, pop_offsets_alias_pop):
        sim_end = self._run_conf["Duration"]
        rep_type = rep_conf["Type"]
        start_time = rep_conf["StartTime"]
        end_time = rep_conf.get("EndTime", sim_end)
        rep_dt = rep_conf["Dt"]
        rep_format = rep_conf["Format"]

        lfp_disabled = not self._circuits.global_manager._lfp_manager._lfp_file
        if rep_type == "lfp" and lfp_disabled:
            logging.error("LFP reports are disabled. Electrodes file might be missing"
                          " or simulator is not set to CoreNEURON")
            return None
        logging.info(
            " * %s (Type: %s, Target: %s, Dt: %f)",
            rep_name,
            rep_type,
            rep_conf["Target"],
            rep_dt
        )

        if rep_format != "SONATA":
            if MPI.rank == 0:
                logging.error("Unsupported report format: '%s'. "
                              "Use 'SONATA' instead.", rep_format)
                return None

        if Nd.t > 0:
            start_time += Nd.t
            end_time += Nd.t
        if end_time > sim_end:
            end_time = sim_end
        if start_time > end_time:
            if MPI.rank == 0:
                logging.error("Report/Sim End-time (%s) before Start (%g).",
                                end_time, start_time)
            return None

        if rep_dt < Nd.dt:
            if MPI.rank == 0:
                logging.error("Invalid report dt %f < %f simulation dt", rep_dt, Nd.dt)
            return None

        electrode = self._elec_manager.getElectrode(rep_conf["Electrode"]) \
            if SimConfig.use_neuron and "Electrode" in rep_conf else None
        rep_target = TargetSpec(rep_conf["Target"])
        population_name = (rep_target.population or self._target_spec.population
                            or self._default_population)
        log_verbose("Report on Population: %s, Target: %s", population_name, rep_target.name)

        target_population = rep_target.population or self._target_spec.population
        pop_offsets, alias_pop = pop_offsets_alias_pop
        offset = pop_offsets[target_population] if target_population in pop_offsets \
                    else pop_offsets[alias_pop[target_population]]
        if offset > 0:  # dont reset if not needed (recent hoc API)
            target.set_offset(offset)

        report_on = rep_conf["ReportOn"]
        return self.ReportParams(
            os.path.basename(rep_conf.get("FileName", rep_name)),
            rep_type,  # rep type is case sensitive !!
            report_on,
            rep_conf["Unit"],
            rep_format,
            rep_dt,
            start_time,
            end_time,
            SimConfig.output_root,
            electrode,
            Nd.String(rep_conf["Scaling"]) if "Scaling" in rep_conf else None,
            rep_conf.get("ISC", "")
        )

    #
    def _report_write_coreneuron_config(self, rep_name, rep_conf, target, rep_params):
        target_spec = TargetSpec(rep_conf["Target"])
        corenrn_target = target

        # 0=Compartment, 1=Cell, Section { 2=Soma, 3=Axon, 4=Dendrite, 5=Apical }
        target_type = target.isCellTarget()
        compartment_offset = 0

        # Note: CoreNEURON does not support more than one Section in a Compartment target
        if target.isCompartmentTarget():
            for activeTarget in target.subtargets:
                if activeTarget.isSectionTarget():
                    # Ensure only one Section (i.e., number of subtargets is also 1)
                    if target.subtargets.count() > 1:
                        logging.error("Report '%s': only a single Section is allowed in a "
                                      "Compartment target", rep_name)
                        return False

                    # If we reach this point, update original target and offset
                    corenrn_target = activeTarget
                    compartment_offset = 4

        if corenrn_target.isSectionTarget():
            if (corenrn_target.subtargets.count() != 1
                    or not corenrn_target.subtargets[0].isCellTarget()):
                logging.error("Report '%s': only a single Cell subtarget is allowed in a "
                              "Section target", rep_name)
                return False

            section_type = corenrn_target.targetSubsets[0].s

            if section_type == "soma":
                target_type = 2 + compartment_offset
            elif section_type == "axon":
                target_type = 3 + compartment_offset
            elif section_type == "dend":
                target_type = 4 + compartment_offset
            elif section_type == "apic":
                target_type = 5 + compartment_offset
            else:
                target_type = 0

        # for sonata config, compute target_type from user inputs
        if "Sections" in rep_conf and "Compartments" in rep_conf:
            def _compute_corenrn_target_type(section_type, compartment_type):
                sections = ["all", "soma", "axon", "dend", "apic"]
                compartments = ["center", "all"]
                if section_type not in sections:
                    raise ConfigurationError(f"Report: invalid section type '{section_type}'")
                if compartment_type not in compartments:
                    raise ConfigurationError(f"Report: invalid compartment type {compartment_type}")
                if section_type == "all":  # for "all sections", support only target_type=0
                    return 0
                return sections.index(section_type)+1+4*compartments.index(compartment_type)

            section_type = rep_conf.get("Sections")
            compartment_type = rep_conf.get("Compartments")
            target_type = _compute_corenrn_target_type(section_type, compartment_type)

        reporton_comma_separated = ",".join(rep_params.report_on.split())
        core_report_params = (
            (os.path.basename(rep_conf.get("FileName", rep_name)),
                target_spec.name, rep_params.rep_type, reporton_comma_separated)
            + rep_params[3:5] + (target_type,) + rep_params[5:8]
            + (target.get_gids(), SimConfig.corenrn_buff_size)
        )
        CoreConfig.write_report_config(*core_report_params)
        return True

    def _report_setup(self, report, rep_conf, target, rep_type):
        # TODO: Move to Cell Distributor and avoid inner loop conditions
        global_manager = self._circuits.global_manager

        if rep_type not in ("compartment", "Summation", "Synapse", "lfp"):
            logging.warning("Unsupported report type: %s.", rep_type)
            return  # Nothing to do

        # Go through the target members, one cell at a time. We give a cell reference
        # For summation targets - check if we were given a Cell target because we really
        # want all points of the cell which will ultimately be collapsed to a single
        # value on the soma. Otherwise, get target points as normal.
        sections = rep_conf.get("Sections")
        compartments = rep_conf.get("Compartments")
        is_cell_target = target.isCellTarget(sections=sections,
                                             compartments=compartments)
        # In case of summation in the soma, we need all points anyway
        if is_cell_target and rep_type == "Summation":
            sections = "all"
            compartments = "all"
        points = self._target_manager.get_target_points(target, global_manager,
                                                        sections=sections,
                                                        compartments=compartments)
        for point in points:
            gid = point.gid
            pop_name, pop_offset = global_manager.getPopulationInfo(gid)
            cell = global_manager.get_cellref(gid)
            spgid = global_manager.getSpGid(gid)

            # may need to take different actions based on report type
            if rep_type == "compartment":
                report.addCompartmentReport(
                    cell, point, spgid, SimConfig.use_coreneuron, pop_name, pop_offset)
            elif rep_type == "Summation":
                report.addSummationReport(
                    cell, point, is_cell_target, spgid, SimConfig.use_coreneuron,
                    pop_name, pop_offset)
            elif rep_type == "Synapse":
                report.addSynapseReport(
                    cell, point, spgid, SimConfig.use_coreneuron, pop_name, pop_offset)

    def _reports_init(self, pop_offsets_alias):
        pop_offsets = pop_offsets_alias[0]

        if SimConfig.use_coreneuron:
            # write spike populations
            if hasattr(CoreConfig, "write_population_count"):
                # Do not count populations with None pop_name
                pop_count = (len(pop_offsets) - 1 if None in pop_offsets else len(pop_offsets))
                CoreConfig.write_population_count(pop_count)
            for pop_name, offset in pop_offsets.items():
                if pop_name is not None:
                    CoreConfig.write_spike_population(pop_name or "All", offset)
            spike_path = self._run_conf.get("SpikesFile")
            if spike_path is not None:
                # Get only the spike file name
                file_name = spike_path.split('/')[-1]
            else:
                file_name = "out.h5"
            CoreConfig.write_spike_filename(file_name)
        else:
            # Report Buffer Size hint in MB.
            reporting_buffer_size = self._run_conf.get("ReportingBufferSize")
            if reporting_buffer_size is not None:
                self._sonatareport_helper.set_max_buffer_size_hint(reporting_buffer_size)

            # once all reports are created, we finalize the communicator for any reports
            self._sonatareport_helper.make_comm()
            self._sonatareport_helper.prepare_datasets()

    # -
    @mpi_no_errors
    def execute_neuron_configures(self):
        """Iterate over any NeuronConfigure blocks from the BlueConfig.
        These are simple hoc statements that can be executed with minimal substitutions
        """
        printed_warning = False

        for config in SimConfig.configures.values():
            if not printed_warning:
                logging.warning("NeuronConfigure block is deprecated")
                printed_warning = True

            target_name = config.get("Target").s
            configure_str = config.get("Configure").s
            log_verbose("Apply configuration \"%s\" on target %s",
                        config.get("Configure").s, target_name)

            points = self._target_manager.get_target_points(target_name,
                                                            self._circuits.base_cell_manager)
            # iterate the pointlist and execute the command on the section
            for tpoint_list in points:
                for sec_i, sc in enumerate(tpoint_list.sclst):
                    if not sc.exists():
                        continue
                    x = tpoint_list.x[sec_i]
                    tstr = configure_str.replace("%s", Nd.secname(sec=sc.sec))
                    tstr = tstr.replace("%g", "%g" % x)
                    Nd.execute1(tstr, sec=sc.sec)

    # -
    @mpi_no_errors
    def sim_init(self, corenrn_gen=None, **sim_opts):
        """Finalize the model and prepare to run simulation.

        After finalizing the model, will eventually write coreneuron config
        and initialize the neuron simulation if applicable.

        Args:
            corenrn_gen: Whether to generate coreneuron config. Default: None (if required)
            sim_opts - override _finalize_model options. E.g. spike_compress
        """
        if self._sim_ready:
            return self._pc

        if not len(self._circuits.all_node_managers()):
            raise RuntimeError("No CellDistributor was initialized. Please create a circuit.")

        self._finalize_model(**sim_opts)

        if corenrn_gen is None:
            corenrn_gen = SimConfig.use_coreneuron
        if corenrn_gen:
            self._sim_corenrn_write_config()

        if SimConfig.use_neuron:
            self._sim_init_neuron()

        if ospath.isfile("debug_gids.txt"):
            self.dump_circuit_config()
        if self._pr_cell_gid:
            self.dump_cell_config()

        self._sim_ready = True
        return self._pc

    # -
    @mpi_no_errors
    @timeit(name="Model Finalized")
    def _finalize_model(self, spike_compress=3):
        """Set up simulation run parameters and initialization.

        Handles setup_transfer, spike_compress, _record_spikes, stdinit, forward_skip, timeout
        Args:
            spike_compress: The spike_compress() parameters (tuple or int)
        """
        logging.info("Preparing to run simulation...")
        is_save_state = SimConfig.save or SimConfig.restore
        self._pc.setup_transfer()

        if spike_compress and not is_save_state and not self._is_ngv_run:
            # multisend 13 is combination of multisend(1) + two_phase(8) + two_intervals(4)
            # to activate set spike_compress=(0, 0, 13)
            if not isinstance(spike_compress, tuple):
                spike_compress = (spike_compress, 1, 0)
            self._pc.spike_compress(*spike_compress)

        # LFP calculation requires WholeCell balancing and extracellular mechanism.
        # This is incompatible with efficient caching atm AND Incompatible with mcd & Glut
        if not self._is_ngv_run:
            Nd.cvode.cache_efficient("ElectrodesPath" not in self._run_conf)
        self._pc.set_maxstep(4)
        with timeit(name="stdinit"):
            Nd.stdinit()

        logging.info("Executing actions after stdinit...")
        for mgr in self._circuits.all_node_managers():
            mgr.post_stdinit()

    # -
    def _sim_init_neuron(self):
        # === Neuron specific init ===
        restore_path = SimConfig.restore
        fwd_skip = self._run_conf.get("ForwardSkip", 0)

        if fwd_skip and not restore_path:
            logging.info("Initializing with ForwardSkip %d ms", fwd_skip)
            Nd.t = -1e9
            prev_dt = Nd.dt
            Nd.dt = fwd_skip * 0.1
            for flushIndex in range(10):
                Nd.fadvance()
            Nd.dt = prev_dt
            Nd.t = 0
            Nd.frecord_init()

        # create a spike_id vector which stores the pairs for spikes and timings for
        # every engine
        for cell_manager in self._circuits.all_node_managers():
            if cell_manager.population_name is not None:
                self._spike_populations.append(
                    (cell_manager.population_name, cell_manager.local_nodes.offset))
                self._spike_vecs.append(cell_manager.record_spikes() if cell_manager.record_spikes()
                                        else (Nd.Vector(), Nd.Vector()))

        self._pc.timeout(200)  # increase by 10x

        if restore_path:
            with timeit(name="restoretime"):
                logging.info("Restoring state...")
                self._stim_manager.saveStatePreparation(self._bbss)
                self._stim_manager.reevent()
                self._bbss.vector_play_init()
                self._restart_events()  # On restore the event queue is cleared
                return  # Upon restore sim is ready

    # -
    def _restart_events(self):
        logging.info("Restarting connections events (Replay and Spont Minis)")
        for syn_manager in self._circuits.all_synapse_managers():
            syn_manager.restart_events()

    @contextmanager
    def _coreneuron_ensure_all_ranks_have_gids(self, corenrn_data):
        local_gid_count = sum(len(manager.local_nodes)
                              for manager in self._circuits.all_node_managers())
        if local_gid_count > 0:
            yield
            return

        # create a fake node with a fake population "zzz" to get an unused gid.
        # It only works with a single cycles without enough cells, but in
        #  _instantiate_simulation we made sure we limit the number of cycles
        log_verbose("Creating fake gid for CoreNeuron")
        pop_group = PopulationNodes.get("zzz", create=True)
        fake_gid = pop_group.offset + 1 + MPI.rank
        # Add the fake cell to the base manager
        base_manager = self._circuits.base_cell_manager
        base_manager.load_artificial_cell(fake_gid, CoreConfig.artificial_cell_object)
        yield

        # Nd.registerMapping doesn't work for this artificial cell as somatic attr is
        # missing, so create a dummy mapping file manually, required for reporting
        cur_files = glob.glob("%s/*_3.dat" % corenrn_data)
        example_mapfile = cur_files[0]
        with open(example_mapfile, 'rb') as f_mapfile:
            # read the version from the existing mapping file generated by coreneuron
            coredata_version = f_mapfile.readline().rstrip().decode("ascii")
        mapping_file = ospath.join(corenrn_data, "%d_3.dat" % fake_gid)
        if not ospath.isfile(mapping_file):
            with open(mapping_file, "w") as dummyfile:
                dummyfile.write("%s\n0\n" % coredata_version)

    # -
    def _sim_corenrn_configure_datadir(self, corenrn_restore):
        corenrn_datadir = SimConfig.coreneuron_datadir
        corenrn_datadir_shm = SHMUtil.get_datadir_shm(corenrn_datadir)

        # Clean-up any previous simulations in the same output directory
        if self._cycle_i == 0 and corenrn_datadir_shm:
            subprocess.call(['/bin/rm', '-rf', corenrn_datadir_shm])

        # Ensure that we have a folder in /dev/shm (i.e., 'SHMDIR' ENV variable)
        if SimConfig.cli_options.enable_shm and not corenrn_datadir_shm:
            logging.warning("Unknown SHM directory for model file transfer in CoreNEURON.")
        # Try to configure the /dev/shm folder as the output directory for the files
        elif self._cycle_i == 0 and not corenrn_restore and \
                (SimConfig.cli_options.enable_shm and SimConfig.delete_corenrn_data):
            # Check for the available memory in /dev/shm and estimate the RSS by multiplying
            # the number of cycles in the multi-step model build with an approximate factor
            mem_avail = SHMUtil.get_mem_avail()
            shm_avail = SHMUtil.get_shm_avail()
            initial_rss = self._initial_rss
            current_rss = SHMUtil.get_node_rss()
            factor = SHMUtil.get_shm_factor()
            rss_diff = (current_rss - initial_rss) if initial_rss < current_rss else current_rss
            rss_req = int(rss_diff * self._n_cycles * factor)  # 'rss_diff' prevents <0 estimates

            # Sync condition value with all ranks to ensure that all of them can use /dev/shm
            shm_possible = (rss_req < shm_avail) and (rss_req < mem_avail)
            if MPI.allreduce(int(shm_possible), MPI.SUM) == MPI.size:
                logging.info("SHM file transfer mode for CoreNEURON enabled")

                # Create SHM folder and links to GPFS for the global data structures
                os.makedirs(corenrn_datadir_shm, exist_ok=True)

                # Important: These three files must be available on every node, as they are shared
                #            across all of the processes. The trick here is to fool NEURON into
                #            thinking that the files are written in /dev/shm, but they are actually
                #            written on GPFS. The workflow is identical, meaning that rank 0 writes
                #            the content and every other rank reads it afterwards in CoreNEURON.
                for filename in {"bbcore_mech.dat", "files.dat", "globals.dat"}:
                    path = os.path.join(corenrn_datadir, filename)
                    path_shm = os.path.join(corenrn_datadir_shm, filename)

                    try:
                        os.close(os.open(path, os.O_CREAT))
                        os.symlink(path, path_shm)
                    except FileExistsError:
                        pass  # Ignore if other process has already created it

                # Update the flag to confirm the configuration
                self._shm_enabled = True
            else:
                logging.warning("Unable to utilize SHM for model file transfer in CoreNEURON. "
                                "Increase the number of nodes to reduce the memory footprint "
                                "(Current use node: %d MB / SHM Limit: %d MB / Mem. Limit: %d MB)",
                                (rss_req >> 20), (shm_avail >> 20), (mem_avail >> 20))

        return corenrn_datadir if not self._shm_enabled else corenrn_datadir_shm

    # -
    @timeit(name="corewrite")
    def _sim_corenrn_write_config(self, corenrn_restore=False):
        log_stage("Dataset generation for CoreNEURON")
        CoreConfig.datadir = self._sim_corenrn_configure_datadir(corenrn_restore)
        fwd_skip = self._run_conf.get("ForwardSkip", 0) if not corenrn_restore else 0

        if not corenrn_restore:
            Nd.registerMapping(self._circuits.global_manager)
            with self._coreneuron_ensure_all_ranks_have_gids(CoreConfig.datadir):
                self._pc.nrnbbcore_write(CoreConfig.datadir)
                MPI.barrier()  # wait for all ranks to finish corenrn data generation

        CoreConfig.write_sim_config(
            Nd.tstop,
            Nd.dt,
            fwd_skip,
            self._pr_cell_gid or -1,
            getattr(SimConfig, 'celsius', 34.0),
            getattr(SimConfig, 'v_init', -65.0),
            self._core_replay_file,
            SimConfig.rng_info.getGlobalSeed(),
            int(SimConfig.cli_options.model_stats),
            int(self._run_conf["EnableReports"])
        )
        # Wait for rank0 to write the sim config file
        MPI.barrier()

        logging.info(" => Dataset written to '{}'".format(CoreConfig.datadir))

    # -
    def run_all(self):
        """Run the whole simulation according to BlueConfig
        """
        if not self._sim_ready:
            self.sim_init()

        timings = None
        if SimConfig.use_neuron:
            timings = self._run_neuron()
            if not SimConfig.is_sonata_config:
                self.spike2file("out.dat")
            self.sonata_spikes()
        if SimConfig.use_coreneuron:
            print_mem_usage()
            self.clear_model(avoid_clearing_queues=False)
            self._run_coreneuron()
            if not SimConfig.is_sonata_config:
                self.adapt_spikes("out.dat")
        return timings

    # -
    @return_neuron_timings
    def _run_neuron(self):
        Nd.load_hoc("ShowProgress")  # TODO: Drop this
        _ = Nd.ShowProgress(Nd.cvode, MPI.rank)
        self.solve()
        logging.info("Simulation finished.")

    # -
    def _run_coreneuron(self):
        logging.info("Launching simulation with CoreNEURON")
        CoreConfig.psolve_core(
            getattr(SimConfig, "save", None),
            getattr(SimConfig, "restore", None)
        )

    #
    def _sim_event_handlers(self, tstart, tstop):
        """Create handlers for "in-simulation" events, like activating delayed
        connections, execute Save-State, etc
        """
        events = defaultdict(list)  # each key (time) points to a list of handlers

        if SimConfig.save:
            tsave = SimConfig.save_time or SimConfig.tstop  # Consider 0 as the end too!
            save_f = self._create_save_handler(tsave)
            events[tsave].append(save_f)

        event_list = [(t, events[t]) for t in sorted(events)]
        return event_list

    # -
    def _create_save_handler(self, tsave):
        @timeit(name="savetime")
        def save_f():
            logging.info("Saving State... (t=%f)", tsave)
            MPI.barrier()
            self._stim_manager.saveStatePreparation(self._bbss)
            log_verbose("SaveState Initialization Done")

            # If event at the end of the sim we can actually clearModel() before savestate()
            if SimConfig.save_time is None:
                log_verbose("Clearing model prior to final save")
                self._sonatareport_helper.flush()

            self.dump_cell_config()
            # Clear the model after saving state as the pointers are being recorded in reportinglib
            if SimConfig.save_time is None:
                self.clear_model()
            logging.info(" => Save done successfully")

        return save_f

    # -
    @mpi_no_errors
    @timeit(name="psolve")
    def solve(self, tstop=None):
        """Call solver with a given stop time (default: whole interval).
        Be sure to have sim_init()'d the simulation beforehand
        """
        if not self._sim_ready:
            raise ConfigurationError("Initialize simulation first")

        tstart = Nd.t
        tstop = tstop or Nd.tstop
        event_list = self._sim_event_handlers(tstart, tstop)

        # NOTE: _psolve_loop is called among events in order to eventually split long
        # simulation blocks, where one or more report flush(es) can happen. It is a simplified
        # design relatively to the original version where the report checkpoint would not happen
        # before the checkpoint timeout (25ms default). However there shouldn't be almost any
        # performance penalty since the simulation is already halted between events.

        logging.info("Running simulation until t=%d ms", tstop)
        t = tstart  # default if there are no events
        for t, events in event_list:
            self._psolve_loop(t)
            for event in events:
                event()
            self.dump_cell_config()  # Respect prCellGid on every state change
        # Run until the end
        if t < tstop:
            self._psolve_loop(tstop)
            self.dump_cell_config()

        # Final flush
        self._sonatareport_helper.flush()

    # psolve_loop: There was an issue where MPI collective routines for reporting and spike exchange
    # are mixed such that some cpus are blocked waiting to complete reporting while others to
    # finish spike exchange. As a work-around, periodically halt simulation and flush reports
    # Default is 25 ms / cycle
    def _psolve_loop(self, tstop):
        cur_t = round(Nd.t, 2)  # fp innnacuracies could lead to infinitesimal loops
        buffer_t = SimConfig.buffer_time
        for _ in range(math.ceil((tstop - cur_t) / buffer_t)):
            next_flush = min(tstop, cur_t + buffer_t)
            self._pc.psolve(next_flush)
            cur_t = next_flush
        Nd.t = cur_t

    # -
    @mpi_no_errors
    def clear_model(self, avoid_creating_objs=False, avoid_clearing_queues=True):
        """Clears appropriate lists and other stored references.
        For use with intrinsic load balancing. After creating and evaluating the network using
        round robin distribution, we want to clear the cells and synapses in order to have a
        clean slate on which to instantiate the balanced cells.
        """
        if not self._target_manager.parser:
            # Target parser is the ground block. If not there model is clear
            return

        logging.info("Clearing model")
        self._pc.gid_clear()
        self._target_manager.clear_simulation_data()

        if not avoid_creating_objs:
            if SimConfig.use_neuron:
                if self._sonatareport_helper:
                    self._sonatareport_helper.clear()

        Node.__init__(self, None, None)  # Reset vars

        # Clear BBSaveState
        self._bbss.ignore()

        # Shrink ArrayPools holding mechanism's data in NEURON
        pool_shrink()

        # Free event queues in NEURON
        if not avoid_clearing_queues:
            free_event_queues()

        # Garbage collect all Python objects without references
        gc.collect()

        # Finally call malloc_trim to return all the freed pages back to the OS
        trim_memory()
        print_mem_usage()

    # -------------------------------------------------------------------------
    #  output
    # -------------------------------------------------------------------------

    @run_only_rank0
    def adapt_spikes(self, outfile):
        """Prepend /scatter to spikes file after coreneuron sim finishes.
        """
        output_root = SimConfig.output_root
        outfile = ospath.join(output_root, outfile)
        tmpfile = ospath.join(output_root, 'tmp.out')
        with open(outfile, 'r') as f_src:
            with open(tmpfile, 'w') as f_dest:
                f_dest.write("/scatter\n")
                copyfileobj(f_src, f_dest)
        move(tmpfile, outfile)

    @mpi_no_errors
    def spike2file(self, outfile):
        """ Write the spike events that occured on each node into a single output file.
        Nodes will write in order, one after the other.
        """
        logging.info("Writing spikes to %s", outfile)
        output_root = SimConfig.output_root
        outfile = ospath.join(output_root, outfile)
        spikevec = Nd.Vector()
        idvec = Nd.Vector()
        # merge spike_vecs and id_vecs for SpikeWriter
        for spikes, ids in self._spike_vecs:
            spikevec.append(spikes)
            idvec.append(ids)
        spikewriter = Nd.SpikeWriter()
        spikewriter.write(spikevec, idvec, outfile)

    def sonata_spikes(self):
        """ Write the spike events that occured on each node into a single output SONATA file.
        """
        output_root = SimConfig.output_root
        if hasattr(self._sonatareport_helper, "create_spikefile"):
            # Write spike report for multiple populations if exist
            spike_path = self._run_conf.get("SpikesFile")
            if spike_path is not None:
                # Get only the spike file name
                file_name = spike_path.split('/')[-1]
            else:
                file_name = "out.h5"
            # create a sonata spike file
            self._sonatareport_helper.create_spikefile(output_root, file_name)
            # write spikes per population
            for (population, population_offset), (spikevec, idvec) in zip(self._spike_populations,
                                                                          self._spike_vecs):
                extra_args = (population, population_offset) if population \
                    else ("All", population_offset)
                self._sonatareport_helper.add_spikes_population(spikevec, idvec, *extra_args)
            # write all spike populations
            self._sonatareport_helper.write_spike_populations()
            # close the spike file
            self._sonatareport_helper.close_spikefile()
        else:
            # fallback: write spike report with one single population "ALL"
            logging.warning("Writing spike reports with multiple populations is not supported. "
                            "If needed, please update to a newer version of neurodamus.")
            population = self._target_spec.population or "All"
            extra_args = (population,)
            self._sonatareport_helper.write_spikes(spikevec, idvec, output_root, *extra_args)

    def dump_cell_config(self):
        if not self._pr_cell_gid:
            return
        if self._cell_state_dump_t == Nd.t:   # avoid duplicating output
            return
        log_verbose("Dumping info about cell %d", self._pr_cell_gid)
        simulator = "CoreNeuron" if SimConfig.use_coreneuron else "Neuron"
        self._pc.prcellstate(self._pr_cell_gid, "py_{}_t{}".format(simulator, Nd.t))
        self._cell_state_dump_t = Nd.t

    # -
    def dump_circuit_config(self, suffix="nrn_python"):
        gidvec = self._circuits.base_cell_manager.local_nodes.final_gids()
        log_stage("Dumping cells state")
        suffix += "_t=" + str(Nd.t)

        if not ospath.isfile("debug_gids.txt"):
            logging.info("Debugging all gids")
            gids = gidvec
        else:
            gids = []
            for line in open("debug_gids.txt"):
                line = line.strip()
                if not line:
                    continue
                gid = int(line)
                if gid in gidvec:
                    gids.append(gid)
            if len(gids):
                log_all(logging.INFO, "Rank %d: Debugging %d gids from debug_gids.txt",
                        MPI.rank, len(gids))

        for gid in gids:
            self._pc.prcellstate(gid, suffix)

    # ---------------------------------------------------------------------------
    # Note: This method may be called automatically from Neurodamus.__del__
    #     and therefore it must stay as simple as possible as exceptions are ignored
    def cleanup(self):
        """Have the compute nodes wrap up tasks before exiting.
        """
        # MemUsage constructor will do MPI communications
        print_mem_usage()

        # Coreneuron runs clear the model before starting
        if not SimConfig.use_coreneuron or SimConfig.simulate_model is False:
            self.clear_model(avoid_creating_objs=True)

        if SimConfig.delete_corenrn_data and not SimConfig.dry_run:
            data_folder = SimConfig.coreneuron_datadir
            logging.info("Deleting intermediate data in %s", data_folder)

            with timeit(name="Delete corenrn data"):
                if MPI.rank == 0:
                    if ospath.islink(data_folder):
                        # in restore, coreneuron data is a symbolic link
                        os.unlink(data_folder)
                    else:
                        subprocess.call(['/bin/rm', '-rf', data_folder])
                    os.remove(ospath.join(SimConfig.output_root, "sim.conf"))
                    if self._run_conf["EnableReports"]:
                        os.remove(ospath.join(SimConfig.output_root, "report.conf"))

                # Delete the SHM folder if it was used
                if self._shm_enabled:
                    data_folder_shm = SHMUtil.get_datadir_shm(data_folder)
                    logging.info("Deleting intermediate SHM data in %s", data_folder_shm)
                    subprocess.call(['/bin/rm', '-rf', data_folder_shm])

            MPI.barrier()

        logging.info("Finished")
        TimerManager.timeit_show_stats()


# Helper class
# ------------
class Neurodamus(Node):
    """A high level interface to Neurodamus
    """

    def __init__(
        self, config_file,
        auto_init=True,
        cleanup_atexit=True,
        logging_level=None,
        **user_opts
    ):
        """Creates and initializes a neurodamus run node

        As part of Initiazation it calls:
         * load_targets
         * compute_load_balance
         * Build the circuit (cells, synapses, GJs)
         * Add stimulus & replays
         * Activate reports if requested

        Args:
            config_file: The BlueConfig recipe file
            logging_level: (int) Redefine the global logging level.
                0 - Only warnings / errors
                1 - Info messages (default)
                2 - Verbose
                3 - Debug messages
            cleanup_atexit: (bool) Call cleanup in the destructor
                [for more see: https://bbpteam.epfl.ch/project/issues/browse/BBPBGLIB-976]
            user_opts: Options to Neurodamus overriding BlueConfig
        """
        self._init_ok = False
        self.cleanup_atexit = cleanup_atexit
        enable_reports = not user_opts.pop("disable_reports", False)

        if logging_level is not None:
            GlobalConfig.verbosity = logging_level

        Node.__init__(self, config_file, user_opts)
        # Use the run_conf dict to avoid passing it around
        self._run_conf["EnableReports"] = enable_reports
        self._run_conf["AutoInit"] = auto_init

        if SimConfig.dry_run:
            self.load_targets()
            self.create_cells()
            self.create_synapses()
            self._init_ok = True
            return

        if SimConfig.restore_coreneuron:
            self._coreneuron_restore()
        elif SimConfig.build_model:
            self._instantiate_simulation()

        # In case an exception occurs we must prevent the destructor from cleaning
        self._init_ok = True

        # Remove .SUCCESS file if exists
        self._success_file = SimConfig.config_file + ".SUCCESS"
        self._remove_file(self._success_file)

    # -
    def _build_model(self):
        log_stage("================ CALCULATING LOAD BALANCE ================")
        load_bal = self.compute_load_balance()
        print_mem_usage()

        log_stage("==================== BUILDING CIRCUIT ====================")
        self.create_cells(load_bal)
        self.execute_neuron_configures()
        print_mem_usage()

        # Create connections
        self.create_synapses()
        print_mem_usage()

        log_stage("================ INSTANTIATING SIMULATION ================")
        # Apply replay
        self.enable_replay()
        print_mem_usage()

        if self._run_conf["AutoInit"]:
            self.init()

    # -
    def init(self):
        """Explicitly initialize, allowing users to make last changes before simulation
        """
        log_stage("Creating connections in the simulator")
        base_seed = self._run_conf.get("BaseSeed", 0)  # base seed for synapse RNG
        for syn_manager in self._circuits.all_synapse_managers():
            syn_manager.finalize(base_seed, SimConfig.use_coreneuron)
        print_mem_usage()

        self.enable_stimulus()
        print_mem_usage()
        self.enable_modifications()

        if self._run_conf["EnableReports"]:
            self.enable_reports()
        print_mem_usage()

        self.sim_init()

    # -
    def _merge_filesdat(self, ncycles):
        log_stage("Generating merged CoreNeuron files.dat")
        coreneuron_datadir = SimConfig.coreneuron_datadir
        cn_entries = []
        for i in range(ncycles):
            log_verbose("files_{}.dat".format(i))
            filename = ospath.join(coreneuron_datadir, "files_{}.dat".format(i))
            with open(filename) as fd:
                first_line = fd.readline()
                nlines = int(fd.readline())
                for _ in range(nlines):
                    line = fd.readline()
                    cn_entries.append(line)

        cnfilename = ospath.join(coreneuron_datadir, "files.dat")
        with open(cnfilename, 'w') as cnfile:
            cnfile.write(first_line)
            cnfile.write(str(len(cn_entries)) + '\n')
            cnfile.writelines(cn_entries)

        logging.info(" => {} files merged successfully".format(ncycles))

    # -
    def _coreneuron_restore(self):
        log_stage(" =============== CORENEURON RESTORE ===============")
        self.load_targets()
        self.enable_replay()
        if self._run_conf["EnableReports"]:
            self.enable_reports()
        self._sim_corenrn_write_config(corenrn_restore=True)
        self._sim_ready = True

    # -
    def _instantiate_simulation(self):
        # Keep the initial RSS for the SHM file transfer calculations
        self._initial_rss = SHMUtil.get_node_rss()
        print_mem_usage()

        self.load_targets()

        # Check connection block configuration and raise warnings for overriding parameters
        SimConfig.check_connections_configure(self._target_manager)

        # Check if user wants to build the model in several steps (only for CoreNeuron)
        n_cycles = SimConfig.modelbuilding_steps

        # Without multi-cycle, it's a trivial model build. sub_targets is False
        if n_cycles == 1:
            self._build_model()
            return

        target = self._target_manager.get_target(self._target_spec)
        target_name = self._target_spec.name
        cell_count = target.gid_count()
        logging.info("Simulation target: %s, Cell count: %d", target_name, cell_count)

        if SimConfig.use_coreneuron and cell_count/n_cycles < MPI.size and cell_count > 0:
            # coreneuron with no. ranks >> no. cells
            # need to assign fake gids to artificial cells in empty threads during module building
            # fake gids start from max_gid + 1
            # currently not support engine plugin where target is loaded later
            max_num_cycles = int(cell_count / MPI.size) or 1
            if n_cycles > max_num_cycles:
                logging.warning("Your simulation is using multi-cycle without enough cells.\n"
                                "  => Number of cycles has been automatically set to the max: %d",
                                max_num_cycles)
                n_cycles = max_num_cycles

        if n_cycles == 1:
            self._build_model()
            return

        logging.info("MULTI-CYCLE RUN: {} Cycles".format(n_cycles))
        TimerManager.archive(archive_name="Before Cycle Loop")

        if SimConfig.is_sonata_config:
            PopulationNodes.freeze_offsets()
            sub_targets = target.generate_subtargets(n_cycles)

            for cycle_i, cur_targets in enumerate(sub_targets):
                logging.info("")
                logging.info("-" * 60)
                log_stage("==> CYCLE {} (OUT OF {})".format(cycle_i + 1, n_cycles))
                logging.info("-" * 60)

                self.clear_model()

                for cur_target in cur_targets:
                    self._target_manager.register_target(cur_target)
                    pop = list(cur_target.population_names)[0]
                    for circuit in itertools.chain([self._base_circuit],
                                                   self._extra_circuits.values()):
                        tmp_target_spec = TargetSpec(circuit.CircuitTarget)
                        if tmp_target_spec.population == pop:
                            tmp_target_spec.name = cur_target.name
                            circuit.CircuitTarget = str(tmp_target_spec)

                self._cycle_i = cycle_i
                self._build_model()

                # Move generated files aside (to be merged later)
                if MPI.rank == 0:
                    base_filesdat = ospath.join(SimConfig.coreneuron_datadir, 'files')
                    os.rename(base_filesdat + '.dat', base_filesdat + "_{}.dat".format(cycle_i))
                # Archive timers for this cycle
                TimerManager.archive(archive_name="Cycle Run {:d}".format(cycle_i + 1))

        else:
            self._multi_cycle_run_blueconfig_setting(n_cycles)

        if MPI.rank == 0:
            self._merge_filesdat(n_cycles)

    def _multi_cycle_run_blueconfig_setting(self, n_cycles):
        """
            Running multi cycle model buildings for blueconfig settings,
            Different from sonata setting because for blueconfig target is set per circuit
            thus can be different.
            This step will be deprecated once the migration to SONATA is complete.
        """
        sub_targets = defaultdict(list)
        for circuit in itertools.chain([self._base_circuit], self._extra_circuits.values()):
            if not circuit.CircuitPath:
                continue
            if circuit.CircuitTarget is None:
                raise ConfigurationError("Multi building steps requires a circuit target "
                                         "for circuit %s" % circuit._name)

            target_spec = TargetSpec(circuit.CircuitTarget)
            target = self._target_manager.get_target(target_spec)
            circuit_subtargets = target.generate_subtargets(n_cycles)
            sub_targets[circuit._name or "Base"] = circuit_subtargets

        for cycle_i in range(n_cycles):
            logging.info("")
            logging.info("-" * 60)
            log_stage("==> CYCLE {} (OUT OF {})".format(cycle_i + 1, n_cycles))
            logging.info("-" * 60)

            self.clear_model()

            for circuit_name, circuit_subtargets in sub_targets.items():
                cur_target = circuit_subtargets[cycle_i]
                cur_target.name += "_" + circuit_name
                self._target_manager.register_target(cur_target)
                circuit = self._extra_circuits.get(circuit_name, self._base_circuit)
                tmp_target_spec = TargetSpec(circuit.CircuitTarget)
                tmp_target_spec.name = cur_target.name
                circuit.CircuitTarget = str(tmp_target_spec)

            self._cycle_i = cycle_i
            self._build_model()

            # Move generated files aside (to be merged later)
            if MPI.rank == 0:
                base_filesdat = ospath.join(SimConfig.coreneuron_datadir, 'files')
                os.rename(base_filesdat + '.dat', base_filesdat + "_{}.dat".format(cycle_i))
            # Archive timers for this cycle
            TimerManager.archive(archive_name="Cycle Run {:d}".format(cycle_i + 1))

    # -
    @timeit(name="finished Run")
    def run(self):
        """Prepares and launches the simulation according to the loaded config.
        If '--only-build-model' option is set, simulation is skipped.
        """
        if SimConfig.dry_run:
            log_stage("============= DRY RUN (SKIP SIMULATION) =============")
            self._dry_run_stats.display_total()
            self._dry_run_stats.display_node_suggestions()
            return
        if not SimConfig.simulate_model:
            self.sim_init()
            log_stage("======== [SKIPPED] SIMULATION (MODEL BUILD ONLY) ========")
        elif not SimConfig.build_model:
            log_stage("============= SIMULATION (SKIP MODEL BUILD) =============")
            self._run_coreneuron()
        else:
            log_stage("======================= SIMULATION =======================")
            self.run_all()
        # Create SUCCESS file if the simulation finishes successfully
        self._touch_file(self._success_file)
        logging.info("Creating .SUCCESS file: '%s'", self._success_file)

    def __del__(self):
        if self._init_ok and self.cleanup_atexit:
            self.cleanup()

    @run_only_rank0
    def _remove_file(self, file_name):
        import contextlib
        with contextlib.suppress(FileNotFoundError):
            os.remove(file_name)

    @run_only_rank0
    def _touch_file(self, file_name):
        with open(file_name, 'a'):
            os.utime(file_name, None)



"""
Neurodamus is a software for handling neuronal simulation using neuron.

Copyright (c) 2018 Blue Brain Project, EPFL.
All rights reserved
"""
from neurodamus import Neurodamus
from neurodamus.cell_distributor import CellDistributor
from neurodamus.connection_manager import _ConnectionManagerBase
from neurodamus.connection import Connection
from line_profiler import LineProfiler
from neurodamus.core import MPI
# import cProfile

if __name__ == "__main__":
    if MPI.rank > 0:
        Neurodamus('BlueConfig')
    else:
        # Neurodamus('BlueConfig').run()
        # cProfile.run("nd = Neurodamus('BlueConfig')")
        lp = LineProfiler()
        lp.add_function(CellDistributor._setup)
        lp.add_function(CellDistributor.finalize)
        lp.add_function(Connection._create_synapse)
        lp.add_function(Connection.apply_configuration)
        lp.add_function(Connection.finalize)
        lp.add_function(_ConnectionManagerBase.group_connect)
        lp.add_function(_ConnectionManagerBase.get_synapse_parameters)
        with lp:
            Neurodamus('BlueConfig')
        lp.print_stats(open("line_stats.txt", "w"))



# flake8: noqa
"""
    neurodamus.core
    ---------------

    The neurodamus.core package implements several helper modules for building circuits
    with Neuron.
    They can be seen as a High-Level Neuron API, and several examples are found under `examples`.
"""

from __future__ import absolute_import
from ._engine import EngineBase
from ._neuron import Neuron
from ._mpi import MPI, OtherRankError
from ._neurodamus import NeurodamusCore
from ._utils import *
from .cell import Cell
from .stimuli import CurrentSource, ConductanceSource



"""
Collection of Cell Readers from different sources (Pure HDF5, SynTool...)
"""
from __future__ import absolute_import
import logging
import numpy as np
import libsonata
from collections import defaultdict
from os import path as ospath

from ..core import NeurodamusCore as Nd
from ..core.configuration import ConfigurationError, SimConfig
from ..core import run_only_rank0
from ..metype import METypeManager, METypeItem
from ..utils import compat
from ..utils.logging import log_verbose

EMPTY_GIDVEC = np.empty(0, dtype="uint32")


class CellReaderError(Exception):
    pass


def _ncs_get_total(ncs_f):
    # first lines might be comments. Skip '#'
    for tstr in ncs_f:
        if not tstr.strip().startswith("#"):
            break
    # should have "Cells x"
    try:
        total_circuit_cells = int(tstr.strip().split()[1])
    except IndexError:
        raise CellReaderError("NCS file contains invalid entry: " + tstr)
    return total_circuit_cells


def _ncs_get_cells(ncs_f):
    ncs_f.readline()  # skip the '{'

    for cell_i, line in enumerate(ncs_f):
        line = line.strip()
        if line == "}":
            break
        parts = line.split()
        assert len(parts) >= 5, "Error in ncs line " + line
        _gid = int(parts[0][1:])
        metype = parts[4]
        yield cell_i, _gid, metype


def split_round_robin(all_gids, stride=1, stride_offset=0, total_cells=None):
    """ Splits a numpy ndarray[uint32] round-robin.
    If the array is None generates new arrays based on the nr of total cells
    """
    if all_gids is not None:
        gidvec = all_gids[stride_offset::stride] if stride > 1 else all_gids
        gidvec.sort()
    else:
        assert total_cells, "split_round_robin: total_cells required without gids"
        cell_i = stride_offset + 1  # gids start from 1
        gidvec = np.arange(cell_i, total_cells + 1, stride, dtype="uint32")
    return gidvec


def dry_run_distribution(gid_metype_bundle, stride=1, stride_offset=0, total_cells=None):
    """ Distribute gid in metype bundles for dry run.

    The principle is the following: all gids with the same metype
    have to be assigned to the same rank. This function receives
    a list of list of gids, each sublist containing gids of the same
    metype. The gid_metype_bundle list of lists is generated by the
    retrieve_unique_metype function. This function performs a
    round robin distribution of the inner lists, i.e. it returns
    a list of gids that are sequentially in the same metype.
    The return is a flattened numpy array of gids that shall be
    instantiated on the same rank.

    Example:
        gid_metype_bundle = [[1, 2, 3], [4, 5, 6], [7, 8, 9], [10]]

        stride = 2
        stride_offset = 0
        return = [1, 2, 3, 7, 8, 9]

        stride = 2
        stride_offset = 1
        return = [4, 5, 6, 10]

    Args:
        gid_metype_bundle: list of lists of gids to be distributed
        mpi_size: MPI size
        mpi_rank: MPI rank
        total_cells: total number of cells in the circuit
    Returns:
        A numpy array of gids that are sequentially in the same metype
    """
    if not gid_metype_bundle:
        return EMPTY_GIDVEC

    # if mpi_size is 1, return all gids flattened
    if stride == 1:
        return np.concatenate(gid_metype_bundle)
    else:
        groups = gid_metype_bundle[stride_offset::stride]
        return np.concatenate(groups) if groups else EMPTY_GIDVEC


def load_ncs(circuit_conf, all_gids, stride=1, stride_offset=0):
    """ Obtain the gids and the metypes for cells in the base circuit.

    Args:
        circuit_conf: the Run secgion from the configuration
        all_gids: The cells ids to be loaded. If it's None then all the cells shall be loaded
        stride: If distribution is desired stride can be set to a value > 1
        stride_offset: When using distribution, the offset to be read within the stride
    Returns:
        A tuple of (gids, metypes, total_ncs_cells)
    """
    gids = compat.Vector("I")
    gid2mefile = {}
    # NCS is historically under nrnPath which nowadays is a file path
    ncs_dir = circuit_conf.nrnPath
    if ospath.isfile(ncs_dir):
        ncs_dir = ospath.dirname(ncs_dir)
    ncs_path = ospath.join(ncs_dir, "start.ncs")

    with open(ncs_path, "r") as ncs_f:
        total_ncs_cells = _ncs_get_total(ncs_f)
        if all_gids is None:
            for cellIndex, gid, metype in _ncs_get_cells(ncs_f):
                if cellIndex % stride == stride_offset:
                    gids.append(gid)
                    gid2mefile[gid] = metype
        else:
            # Index desired cells
            gid2mefile = {int(gid): None for i, gid in enumerate(all_gids)
                          if i % stride == stride_offset}
            gids.extend(gid2mefile.keys())
            assigned_metypes = 0
            for cellIndex, gid, metype in _ncs_get_cells(ncs_f):
                if gid in gid2mefile:
                    gid2mefile[gid] = metype
                    assigned_metypes += 1
            if assigned_metypes < len(gid2mefile):
                logging.error("start.ncs: found info only for %d cells (out of %d)",
                              assigned_metypes, len(all_gids))
                raise CellReaderError("Target contains invalid circuit cells")
    return gids, gid2mefile, total_ncs_cells


def load_mvd3(circuit_conf, all_gids, stride=1, stride_offset=0):
    """Load cells from MVD3, required for v6 circuits
       reuse load_nodes with mvdtool by default
       if py-mvdtool not installed, use the old h5py loader
    """
    try:
        import mvdtool  # noqa: F401
    except ImportError:
        logging.warning("Cannot import mvdtool to load mvd3, will load with h5py")
        return _load_mvd3_h5py(circuit_conf, all_gids, stride, stride_offset)

    return load_nodes_mvd3(circuit_conf, all_gids, stride, stride_offset)


def _load_mvd3_h5py(circuit_conf, all_gids, stride=1, stride_offset=0):
    """Load cells from MVD3 using h5py
    """
    import h5py  # Can be heavy so loaded on demand
    pth = ospath.join(circuit_conf.CircuitPath, "circuit.mvd3")
    mvd = h5py.File(pth, 'r')

    mecombo_ds = mvd["/cells/properties/me_combo"]
    total_mvd_cells = len(mecombo_ds)

    gidvec = split_round_robin(all_gids, stride, stride_offset, total_mvd_cells)

    if not len(gidvec):
        # Not enough cells to give this rank a few
        return compat.Vector('I'), METypeManager(), total_mvd_cells

    # Indexes are 0-based, and cant be numpy
    indexes = compat.Vector("I", gidvec - 1)

    morph_ids = mvd["/cells/properties/morphology"][indexes]
    combo_ids = mvd["/cells/properties/me_combo"][indexes]
    morpho_ds = mvd["/library/morphology"]
    morpho_names = [str(morpho_ds[i]) for i in morph_ids]
    combo_ds = mvd["/library/me_combo"]
    combo_names = [str(combo_ds[i]) for i in combo_ids]

    # We require gidvec as compat.Vector
    gidvec = compat.Vector("I", gidvec)

    # now we can take the combo file and get the emodel + additional info
    combo_file = circuit_conf.MEComboInfoFile
    me_manager = load_combo_metypes(combo_file, gidvec, combo_names, morpho_names)

    return gidvec, me_manager, total_mvd_cells


def load_nodes_mvd3(circuit_conf, all_gids, stride=1, stride_offset=0):
    """Load cells from MVD3 file.
       node_population can be provided by load_sonata() and can be None. False to auto-detect
    """
    try:
        import mvdtool
    except ImportError:
        raise ConfigurationError("load_nodes: mvdtool is not available. Please install")
    pth = circuit_conf.CellLibraryFile
    assert pth.endswith('.mvd3'), "CellLibraryFile must be a mvd3 file"
    if not ospath.isfile(pth):
        pth = ospath.join(circuit_conf.CircuitPath, pth)
    if not ospath.isfile(pth):
        raise ConfigurationError("Could not find Nodes: " + circuit_conf.CellLibraryFile)

    node_reader = mvdtool.open(pth)
    combo_file = circuit_conf.MEComboInfoFile

    if not combo_file:
        logging.warning("Missing BlueConfig field 'MEComboInfoFile' which has gid:mtype:emodel")
    else:
        node_reader.open_combo_tsv(combo_file)

    total_cells = len(node_reader)

    meinfo = METypeManager()

    if SimConfig.dry_run:
        raise Exception("Dry run is not supported for MVD3")
    else:
        gidvec = split_round_robin(all_gids, stride, stride_offset, total_cells)

    if not len(gidvec):
        # Not enough cells to give this rank a few
        return gidvec, METypeManager(), total_cells

    fetch_MEinfo(node_reader, gidvec, combo_file, meinfo)

    return gidvec, meinfo, total_cells


def fetch_MEinfo(node_reader, gidvec, combo_file, meinfo):

    indexes = np.sort(gidvec - 1)  # MVDtool requires 0-indexed ids.
    gidvec = indexes + 1  # Return 1-indexed ids
    if len(indexes) < 10:  # Ensure array is not too small (pybind11 #1392)
        indexes = indexes.tolist()

    morpho_names = node_reader.morphologies(indexes)
    mtypes = node_reader.mtypes(indexes)
    emodels = node_reader.emodels(indexes) \
        if combo_file else None  # Rare but we may not need emodels (ngv)
    etypes = node_reader.etypes(indexes) \
        if combo_file else None
    exc_mini_freqs = node_reader.exc_mini_frequencies(indexes) \
        if node_reader.hasMiniFrequencies() else None
    inh_mini_freqs = node_reader.inh_mini_frequencies(indexes) \
        if node_reader.hasMiniFrequencies() else None
    threshold_currents = node_reader.threshold_currents(indexes) \
        if node_reader.hasCurrents() else None
    holding_currents = node_reader.holding_currents(indexes) \
        if node_reader.hasCurrents() else None
    positions = node_reader.positions(indexes)
    rotations = node_reader.rotations(indexes) if node_reader.rotated else None

    meinfo.load_infoNP(gidvec, morpho_names, emodels, mtypes, etypes, threshold_currents,
                       holding_currents, exc_mini_freqs, inh_mini_freqs, positions, rotations)


def load_sonata(circuit_conf, all_gids, stride=1, stride_offset=0, *,
                node_population, load_dynamic_props=(), has_extra_data=False, dry_run_stats=None):
    """
    A reader supporting additional dynamic properties from Sonata files.
    """
    import libsonata
    node_file = circuit_conf.CellLibraryFile
    node_store = libsonata.NodeStorage(node_file)
    node_pop = node_store.open_population(node_population)
    attr_names = node_pop.attribute_names
    dynamics_attr_names = node_pop.dynamics_attribute_names
    total_cells = node_pop.size

    def load_base_info_dry_run():
        CELL_NODE_INFO_LIMIT = 100
        log_verbose("Sonata dry run mode: looking for unique metype instances")
        meinfos = METypeManager()
        # skip_metypes = set(dry_run_stats.metype_memory.keys())
        metype_gids, counts = _retrieve_unique_metypes(node_pop, all_gids)
        dry_run_stats.metype_counts += counts
        dry_run_stats.metype_gids = metype_gids
        gid_metype_bundle = list(metype_gids.values())
        gidvec = dry_run_distribution(gid_metype_bundle, stride, stride_offset, total_cells)

        log_verbose("Loading node attributes... (subset of cells from each metype)")
        for gids in metype_gids.values():
            if not len(gids):
                continue
            gids = gids[:CELL_NODE_INFO_LIMIT]
            node_sel = libsonata.Selection(gids - 1)  # Load 0-based node ids
            morpho_names = node_pop.get_attribute("morphology", node_sel)
            mtypes = node_pop.get_attribute("mtype", node_sel)
            etypes = node_pop.get_attribute("etype", node_sel)
            _model_templates = node_pop.get_attribute("model_template", node_sel)
            emodel_templates = [emodel.removeprefix("hoc:") for emodel in _model_templates]
            meinfos.load_infoNP(gids, morpho_names, emodel_templates, mtypes, etypes)
        return gidvec, meinfos, total_cells

    def load_nodes_base_info():
        if SimConfig.dry_run:
            return load_base_info_dry_run()

        meinfos = METypeManager()
        gidvec = split_round_robin(all_gids, stride, stride_offset, total_cells)

        if not len(gidvec):
            # Not enough cells to give this rank a few
            return gidvec, meinfos, total_cells

        log_verbose("Loading nodes info")
        node_sel = libsonata.Selection(gidvec - 1)  # 0-based node indices
        morpho_names = node_pop.get_attribute("morphology", node_sel)
        mtypes = node_pop.get_attribute("mtype", node_sel)
        try:
            etypes = node_pop.get_attribute("etype", node_sel)
        except libsonata.SonataError:
            logging.warning("etype not found in node population, setting to None")
            etypes = None
        _model_templates = node_pop.get_attribute("model_template", node_sel)
        emodel_templates = [emodel.removeprefix("hoc:") for emodel in _model_templates]
        if set(["exc_mini_frequency", "inh_mini_frequency"]).issubset(attr_names):
            exc_mini_freqs = node_pop.get_attribute("exc_mini_frequency", node_sel)
            inh_mini_freqs = node_pop.get_attribute("inh_mini_frequency", node_sel)
        else:
            exc_mini_freqs = None
            inh_mini_freqs = None
        if set(["threshold_current", "holding_current"]).issubset(dynamics_attr_names):
            threshold_currents = node_pop.get_dynamics_attribute("threshold_current", node_sel)
            holding_currents = node_pop.get_dynamics_attribute("holding_current", node_sel)
        else:
            threshold_currents = None
            holding_currents = None
        positions = np.array([node_pop.get_attribute("x", node_sel),
                              node_pop.get_attribute("y", node_sel),
                              node_pop.get_attribute("z", node_sel)]).T
        rotations = _get_rotations(node_pop, node_sel)

        # For Sonata and new emodel hoc template, we need additional attributes for building metype
        # TODO: validate it's really the emodel_templates var we should pass here, or etype
        add_params_list = None if not has_extra_data \
            else _getNeededAttributes(node_pop, circuit_conf.METypePath, emodel_templates, gidvec-1)

        meinfos.load_infoNP(gidvec, morpho_names, emodel_templates, mtypes, etypes,
                            threshold_currents, holding_currents,
                            exc_mini_freqs, inh_mini_freqs, positions, rotations,
                            add_params_list)
        return gidvec, meinfos, total_cells

    # If dynamic properties are not specified simply return early
    if not load_dynamic_props:
        return load_nodes_base_info()

    # Check properties exist, eventually removing prefix
    def validate_property(prop_name):
        if prop_name.startswith("@dynamics:"):
            actual_prop_name = prop_name[len("@dynamics:"):]  # remove prefix
            if actual_prop_name not in dynamics_attr_names:
                raise Exception('Required Dynamics property %s not present' % prop_name)
        elif prop_name not in attr_names:
            raise Exception('Required extra property %s not present' % prop_name)

    [validate_property(p) for p in load_dynamic_props]

    # All good. Lets start reading!
    gidvec, meinfos, fullsize = load_nodes_base_info()
    node_sel = libsonata.Selection(gidvec - 1)  # 0-based node indices

    for prop_name in load_dynamic_props:
        log_verbose("Loading extra property: %s ", prop_name)
        if prop_name.startswith("@dynamics:"):
            prop_name = prop_name[len("@dynamics:"):]
            prop_data = node_pop.get_dynamics_attribute(prop_name, node_sel)
        else:
            prop_data = node_pop.get_attribute(prop_name, node_sel)
        for gid, val in zip(gidvec, prop_data):
            meinfos[gid].extra_attrs[prop_name] = val

    return gidvec, meinfos, fullsize


def load_combo_metypes(combo_file, gidvec, combo_list, morph_list):
    """ Read file with mecombo info, retaining only those that are local to this node
    Args:
        combo_file: Path to the combo file to read metype info from
        gidvec: local gids to load info about
        combo_list: comboList Combos corresponding to local gids
        morph_list: morphList Morpholgies corresponding to local gids
    """
    if not combo_file:
        logging.error("Missing BlueConfig field 'MEComboInfoFile' which has gid:mtype:emodel.")
        raise ConfigurationError("MEComboInfoFile not specified")

    # Optimization: index combos
    combo_ids = defaultdict(list)
    for i, c in enumerate(combo_list):
        combo_ids[c].append(i)

    log_verbose("Loading emodel+additional info from Combo f %s", combo_file)
    f = open(combo_file)
    next(f)  # Skip Header

    me_manager = METypeManager()
    for tstr in f:
        vals = tstr.strip().split()
        if len(vals) not in (6, 8):
            wmsg = ("Could not parse line %s from MEComboInfoFile %s."
                    "Expecting 6 (hippocampus) or 8 (somatosensory) fields")
            logging.warning(wmsg, tstr, combo_file)

        # metypes may be reused by several cells
        # We create a single item and later assign to each matching gid
        meitem = METypeItem(*vals)
        for i in combo_ids[meitem.combo_name]:
            if morph_list[i] == meitem.morph_name:
                me_manager[int(gidvec[i])] = meitem

    # confirm that all gids have been matched.
    # Otherwise, print combo + morph info to help finding issues
    nerr = 0
    for gid in gidvec:
        gid = int(gid)
        if gid not in me_manager:
            logging.error("MEComboInfoFile: No MEInfo for gid %d", gid)
            nerr += 1
    if nerr > 0:
        logging.error("gidvec: " + str(gidvec))
        logging.error("Memap: " + str(me_manager.gids))
        raise CellReaderError("Errors found during processing of mecombo file. See log")

    return me_manager


def _getNeededAttributes(node_reader, etype_path, emodels, gidvec):
    """
    Read additional attributes required by emodel templates global var <emodel>__NeededAttributes
    Args:
        node_reader: libsonata node population
        etype_path: Location of emodel hoc templates
        emodels: Array of emodel names
        gidvec: Array of 0-based cell gids
    """
    add_params_list = []
    for gid, emodel in zip(gidvec, emodels):
        Nd.h.load_file(ospath.join(etype_path, emodel) + ".hoc")  # hoc doesn't throw
        attr_names = getattr(Nd, emodel + "_NeededAttributes", None)  # format "attr1;attr2;attr3"
        vals = []
        if attr_names is not None:
            vals = [node_reader.get_dynamics_attribute(name, gid) for name in attr_names.split(";")]
        add_params_list.append(vals)
    return add_params_list


def _get_rotations(node_reader, selection):
    """
    Read rotations attributes, returns a double vector of size [N][4] with the rotation quaternions
    in the order (x,y,z,w)
    Args:
        node_reader: libsonata node population
        selection: libsonata selection
    """
    attr_names = node_reader.attribute_names
    if set(["orientation_x", "orientation_y",
            "orientation_z", "orientation_w"]).issubset(attr_names):
        # Preferred way to present the rotation as quaternions
        return np.array([node_reader.get_attribute("orientation_x", selection),
                         node_reader.get_attribute("orientation_y", selection),
                         node_reader.get_attribute("orientation_z", selection),
                         node_reader.get_attribute("orientation_w", selection)]).T
    elif set(["rotation_angle_xaxis",
              "rotation_angle_yaxis",
              "rotation_angle_zaxis"]).intersection(attr_names):
        # Some sonata nodes files use the Euler angle rotations, convert them to quaternions
        from scipy.spatial.transform import Rotation
        angle_x = node_reader.get_attribute("rotation_angle_xaxis", selection) \
            if "rotation_angle_xaxis" in attr_names else 0
        angle_y = node_reader.get_attribute("rotation_angle_yaxis", selection) \
            if "rotation_angle_yaxis" in attr_names else 0
        angle_z = node_reader.get_attribute("rotation_angle_zaxis", selection) \
            if "rotation_angle_yaxis" in attr_names else 0
        euler_rots = np.array([angle_x, angle_y, angle_z]).T
        return Rotation.from_euler("xyz", euler_rots).as_quat()
    else:
        return None


@run_only_rank0
def _retrieve_unique_metypes(node_reader, all_gids, skip_metypes=()) -> dict:
    """
    Find unique mtype+emodel combinations in target to estimate resources in dry run.
    This function returns a list of lists of unique mtype+emodel combinations.
    Each of the inner lists contains gid for the same mtype+emodel combinations.

    Args:
        node_reader: node reader, libsonata only
        all_gids: list of all gids in target
    Returns:
        list of lists of unique mtype+emodel combinations
    """
    gidvec = np.array(all_gids)
    indexes = gidvec - 1
    if len(indexes) < 10:  # Ensure array is not too small (pybind11 #1392)
        indexes = indexes.tolist()

    if isinstance(node_reader, libsonata.NodePopulation):
        etypes = node_reader.get_attribute("etype", libsonata.Selection(indexes))
        mtypes = node_reader.get_attribute("mtype", libsonata.Selection(indexes))
    else:
        raise Exception(f"Reader type {type(node_reader)} incompatible with dry run.")

    gids_per_metype = defaultdict(list)
    count_per_metype = defaultdict(int)
    for gid, mtype, etype in zip(gidvec, mtypes, etypes):
        metype = f"{mtype}-{etype}"
        gids_per_metype[metype].append(gid)
        count_per_metype[metype] += 1

    logging.info("Out of %d cells, found %d unique mtype+emodel combination",
                 len(gidvec), len(gids_per_metype))
    for metype, gid_list in gids_per_metype.items():
        logging.debug("METype: %-20s instances: %-8d gids: %s",
                      metype, len(gid_list), gid_list[:10])

    # If the list is longer than 50, truncate it to 50 elements.
    # If the metype is already computed, skip it
    gid_metype_instantiate = {}
    for metype, gid_list in gids_per_metype.items():
        if metype not in skip_metypes:
            gid_metype_instantiate[metype] = np.array(gid_list, dtype="uint32")
        else:
            log_verbose("Skipping METype '%s' since it's already known", metype)

    return gid_metype_instantiate, count_per_metype



"""
Implementation of Gid Sets with the ability of self offsetting and avoid
global overlapping
"""
from contextlib import contextmanager
import numpy
from ..utils import compat, WeakList
from . import MPI


class PopulationNodes:
    """
    Handle NodeSets belonging to a population. Given that Neuron doesnt
    inherently handle populations, we will have to apply gid offsetting.
    The class stores `NodeSet`s, and makes the required offsetting on-the-fly.

    This class is intended to be internal, since NodeSet instances can be
    freely created but only "apply" for offsetting when registered globally,
    in which case it delegates the processing to PopulationNodes.

    We store internal class-level _global_populations so that offsets are truly
    global, independently of the CellManager
    """

    _global_populations = []
    """Populations which may have offset"""
    _has_base_population = False
    """Select one population to be the first one, no offset"""
    _do_offsetting = True
    """False will freeze offsets to ensure final gids are consistent"""

    def __init__(self, name):
        """Ctor for a group of nodes belonging to the same population.
        It wont probably be used publicly given `get()` is also a factory.
        """
        self.name = name
        self.nodesets = WeakList()  # each population might contain several NodeSet's
        self.max_gid = 0  # maximum raw gid (without offset)
        self.offset = 0

    def _append(self, nodeset):
        self.nodesets.append(nodeset)
        self._update(nodeset)
        return self

    def _update(self, updated_nodeset):
        updated_nodeset._offset = self.offset
        if not self._do_offsetting:
            return
        local_max = max(self.max_gid, updated_nodeset._max_gid)
        max_gid = int(MPI.allreduce(local_max, MPI.MAX))
        if max_gid > self.max_gid:
            self.max_gid = max_gid
            self._update_offsets()

    @classmethod
    def register(cls, population, nodeset, **create_kw):
        return cls.get(population, create=True, **create_kw)._append(nodeset)

    @classmethod
    def freeze_offsets(cls):
        cls._do_offsetting = False

    @classmethod
    def reset(cls):
        cls._global_populations.clear()
        cls._do_offsetting = True

    @classmethod
    def all(cls):
        return cls._global_populations

    @classmethod
    def get(cls, population_name, *, create=False, **create_kw):
        obj = next(filter(lambda x: x.name == population_name, cls._global_populations), None)
        if not obj and create:
            obj = cls.create_pop(population_name, **create_kw)
        return obj

    @classmethod
    def create_pop(cls, population_name, *, is_base_pop=False):
        new_population = cls(population_name)
        if is_base_pop:
            cls._global_populations.insert(0, new_population)
            cls._has_base_population = True
            return new_population

        # Otherwise insert at the end and do sorting
        cls._global_populations.append(new_population)
        base_pop, other_pops = [], cls._global_populations
        if cls._has_base_population:
            base_pop, other_pops = cls._global_populations[0:1], cls._global_populations[1:]
        cls._global_populations = base_pop + sorted(other_pops, key=lambda x: x.name)
        new_population._compute_offset(cls._find_previous(new_population))
        return new_population

    @classmethod
    def _find_previous(cls, cur_pop):
        prev_nodeset = None
        for obj in cls.all():
            if obj is cur_pop:
                return prev_nodeset
            prev_nodeset = obj
        return None

    def _compute_offset(self, prev_gidpop):
        offset = 0
        # This offset is gonna be the offset+max_gid of the previous population, round up
        if prev_gidpop is not None:
            cur_max = prev_gidpop.offset + prev_gidpop.max_gid
            # round up 1000's. GIDs are 1 based: Blocks [1-1000], [1001-2000]
            offset = ((cur_max - 1) // 1000 + 1) * 1000
        self.offset = offset
        # Update individual nodesets
        for nodeset in self.nodesets:  # nodeset is a weakref
            nodeset()._offset = offset

    def _update_offsets(self):
        """Update all global offsets after adding gids"""
        update = False
        prev_gidpop = None
        for gidpop in self.all():
            if gidpop is self:
                update = True
                prev_gidpop = gidpop
                continue
            if update:
                # We are in a subsequent nodeset - re-set offsetting
                gidpop._compute_offset(prev_gidpop)
            prev_gidpop = gidpop

    @classmethod
    @contextmanager
    def offset_freezer(cls):
        cls._do_offsetting = False
        yield
        cls._do_offsetting = True


class _NodeSetBase:
    """
    Common bits between nodesets, so they can be registered globally and get offsets
    """

    def __init__(self, *_, **_kw):
        self._offset  = 0
        self._max_gid = 0  # maximum raw gid (without offset)
        self._population_group = None  # register in a population so gids can be unique

    offset = property(lambda self: self._offset)
    max_gid = property(lambda self: self._max_gid)

    def register_global(self, population_name, is_base_pop=False):
        """ Registers a node set as being part of a population, potentially implying an offsett

        Args:
            population_name: The name of the population these ids belong to
            is_base_population: In case we want this population the be the base, without offset
        """
        self._population_group = PopulationNodes.register(
            population_name, self, is_base_pop=is_base_pop
        )
        return self

    @property
    def population_name(self):
        return self._population_group.name if self._population_group else None

    def _check_update_offsets(self):
        if self._population_group:
            self._population_group._update(self)  # Note: triggers a reduce.

    @classmethod
    def unregister_all(cls):
        PopulationNodes.reset()

    def __len__(self):
        return NotImplemented

    def raw_gids(self):
        return NotImplemented

    def final_gids(self):
        return numpy.add(self.raw_gids(), self._offset, dtype="uint32")

    def intersection(self, _other, _raw_gids=False):
        return NotImplemented

    def intersects(self, other):
        """Check if the current nodeset intersects another

        For nodesets to intersect they must belong to the same population and
        have common gids
        """
        return len(self.intersection(other)) > 0


class NodeSet(_NodeSetBase):
    """A set of nodes. When registered globally offset computation happens
    so that different population's gids dont overlap
    """

    def __init__(self, gids=None, gid_info=None, **metadata):
        """Create a NodeSet.

        Args:
            gids: The gids to handle
            gid_info: a map containing METype information about each cell.
                In v5 and v6 values are METypeItem's

        """
        super().__init__()
        self._gidvec = compat.Vector()  # raw gids
        self._gid_info = {}
        self._metadata = metadata
        if gids is not None:
            self.add_gids(gids, gid_info)

    meta = property(lambda self: self._metadata)

    def add_gids(self, gids, gid_info=None):
        """Add raw gids, recomputing gid offsets as needed"""
        self._gidvec.extend(gids)
        if len(gids) > 0:
            self._max_gid = max(self.max_gid, max(gids))
        if gid_info:
            self._gid_info.update(gid_info)
        self._check_update_offsets()  # check offsets (uses reduce)
        return self

    def extend(self, other):
        return self.add_gids(other._gidvec, other._gid_info)

    def __len__(self):
        return len(self._gidvec)

    def raw_gids(self):
        return numpy.asarray(self._gidvec, dtype="uint32")

    def items(self, final_gid=False):
        offset_add = self._offset if final_gid else 0
        for gid in self._gidvec:
            yield gid + offset_add, self._gid_info.get(gid)

    def intersection(self, other, raw_gids=False):
        """Computes the intersection of two NodeSet's

        For nodesets to intersect they must belong to the same population and
        have common gids. Otherwise an empty list is returned.
        """
        if self.population_name != other.population_name:
            return []
        intersect = numpy.intersect1d(self.raw_gids(), other.raw_gids(), assume_unique=True)
        if raw_gids:
            return intersect
        return numpy.add(intersect, self._offset, dtype="uint32")

    def clear_cell_info(self):
        self._gid_info = None


class SelectionNodeSet(_NodeSetBase):
    """
    A lightweight shim over a `libsonata.Selection` so that gids get offset
    """

    def __init__(self, sonata_selection):
        super().__init__()
        self._selection = sonata_selection
        # Max gid is the end of the last range since we need +1 (1 based)
        self._max_gid = sonata_selection.ranges[-1][1] if sonata_selection.ranges else 0
        self._size = sonata_selection.flat_size

    def __len__(self):
        return self._size

    def raw_gids(self):
        return numpy.add(self._selection.flatten(), 1, dtype="uint32")

    def raw_gids_iter(self):
        for r_start, r_end in self._selection.ranges:
            yield from range(r_start + 1, r_end + 1)

    def final_gids_iter(self):
        for gid in self.raw_gids_iter():
            yield gid + self._offset

    def intersection(self, other: _NodeSetBase, raw_gids=False, _quick_check=False):
        """Computes intersection of two nodesets.
        """
        # NOTE: A _quick_check param can be set to True so that we effectively only check for
        # intersection (True/False) instead of computing the actual intersection (internal).
        if self.population_name != other.population_name:
            return []

        sel2 = getattr(other, "_selection", None)
        if sel2:
            intersect = _ranges_overlap(self._selection.ranges, sel2.ranges, True, _quick_check)
        else:
            # Selection ranges are 0-based. We must bring gids to 0-based
            base_gids = numpy.subtract(other.raw_gids(), 1, dtype="uint32")
            intersect = _ranges_vec_overlap(self._selection.ranges, base_gids, _quick_check)

        if _quick_check:
            return intersect
        if len(intersect):
            if raw_gids:
                # TODO: We should change the return type to be another `SelectionNodeSet`
                # Like that we could still keep ranges internally and have PROPER API to get raw ids
                return numpy.add(intersect, 1, dtype=intersect.dtype)
            return numpy.add(intersect, self.offset + 1, dtype=intersect.dtype)
        return numpy.array([], dtype="uint32")

    def intersects(self, other):
        return self.intersection(other, _quick_check=True)


def _ranges_overlap(ranges1, ranges2, flattened_out=False, quick_check=False, dtype="uint32"):
    """
    Detect overlaps between two lists of ranges.
    This is especially important for nodesets since we can access the ranges in no time
    without the need to flatten and consume GBs of memory

    Args:
        ranges1: The first list of ranges
        ranges2: The second list of ranges
        flattened_out (bool): Whether to return the overlaps as a flat list. Otherwise range list
        quick_check: Whether to short-circuit and return True if any overlap exists
        dtype: The output dtype in case flattened_out is requested [default: "uint32"]
    """

    if not ranges1 or not ranges2:
        return []

    all_ranges = []
    r1_iter = iter(ranges1)
    r2_iter = iter(ranges2)
    r1, r2 = next(r1_iter), next(r2_iter)

    while r1 and r2:
        if r2[0] >= r1[1]:  # r2 past over end r1. Move r1
            r1 = next(r1_iter, None)
            continue
        if r2[1] <= r1[0]:  # r2 before whole range r1. Move r2
            r2 = next(r2_iter, None)
            continue

        # Phew, finally some intersection
        low, high = max(r1[0], r2[0]), min(r1[1], r2[1])
        if low < high:
            if quick_check:
                return True
            all_ranges.append((low, high))

        # Now move the one that still has potential for more overlap without moving the other
        if r2[1] > r1[1]:
            r1 = next(r1_iter, None)
        else:
            r2 = next(r2_iter, None)

    if quick_check:
        return False  # We know it's False as quick_check returns True in the loop
    if not flattened_out:
        return all_ranges
    if not all_ranges:
        return []
    return numpy.concatenate([numpy.arange(*r, dtype=dtype) for r in all_ranges])


def _ranges_vec_overlap(ranges1, vector, quick_check=False):
    """
    Detect overlaps between a list of ranges and a vector of ints
    This is particularly used to know the overlap between a SelectionNodeSet and a list
    of gids, e.g. the list of local gids.

    Args:
        ranges1: The list of ranges
        vector: The array of values to intersect with
        quick_check: Whether to short-circuit and return True if any overlap exists
    """
    if not ranges1 or len(vector) == 0:
        return []
    vector = numpy.asarray(vector)
    all_ranges = []

    for r1 in ranges1:
        if vector[-1] < r1[0]:  # gids before whole range r1
            break
        if vector[0] >= r1[1]:  # r2 past over end r1
            continue
        mask = (r1[0] <= vector) & (vector < r1[1])
        if numpy.any(mask):
            if quick_check:
                return True
            all_ranges.append(vector[mask])

    if quick_check:
        return False
    if not all_ranges:
        return []
    return numpy.concatenate(all_ranges)



# https://bbpteam.epfl.ch/project/spaces/display/BGLIB/Neurodamus
# Copyright 2005-2021 Blue Brain Project, EPFL. All rights reserved.
"""
    Implements coupling artificial stimulus into simulation

    New Stimulus classes must be registered, using the appropriate decorator.
    Also, when instantiated by the framework, __init__ is passed three arguments
    (1) target (2) stim_info: dict (3) cell_manager. Example

    >>> @StimulusManager.register_type
    >>> class ShotNoise:
    >>>
    >>> def __init__(self, target, stim_info: dict, cell_manager):
    >>>     tpoints = target.getPointList(cell_manager)
    >>>     for point in tpoints:
    >>>         gid = point.gid
    >>>         cell = cell_manager.get_cell(gid)

"""

import logging
from .core import NeurodamusCore as Nd
from .utils.logging import log_verbose
from .core.configuration import SimConfig
from .core.stimuli import CurrentSource, ConductanceSource
from .core import random


class StimulusManager:

    """
    A manager for synaptic artificial Stimulus.
    Old stimulus resort to hoc implementation
    """

    _stim_types = {}  # stimulus handled in Python

    def __init__(self, target_manager, elec_manager=None):
        base_seed = SimConfig.run_conf.get("BaseSeed")
        xargs = () if base_seed is None else (base_seed,)
        self._hoc = Nd.StimulusManager(target_manager, elec_manager, *xargs)
        self._target_manager = target_manager
        self._stim_seed = SimConfig.run_conf.get("StimulusSeed")
        self._stimulus = []
        self.reset_helpers()  # reset helpers for multi-cycle builds

    def interpret(self, target_spec, stim_info):
        stim_t = self._stim_types.get(stim_info["Pattern"])
        if self._stim_seed is None and getattr(stim_t, 'IsNoise', False):
            logging.warning("StimulusSeed unset (default %d), "
                            "set explicitly to vary noisy stimuli across runs",
                            SimConfig.rng_info.getStimulusSeed())
        # Get either hoc target or sonata node_set, needed for python and hoc interpret
        # If sonata node_set, internally register the target and add to hoc TargetList
        target = self._target_manager.get_target(target_spec)
        if SimConfig.cli_options.experimental_stims or stim_t and stim_t.IsPythonOnly:
            # New style Stim, in Python
            log_verbose("Using new-gen stimulus")
            cell_manager = self._target_manager.hoc.cellDistributor
            stim = stim_t(target, stim_info, cell_manager)
            self._stimulus.append(stim)
        else:
            # Fallback to hoc stim manager
            self._hoc.interpret(target_spec.name, stim_info.hoc_map)

    def interpret_extracellulars(self, injects, stimuli):
        """Hoc only implementation for extra-cellulars"""
        self._hoc.interpretExtracellulars(injects.hoc_map, stimuli.hoc_map)

    def __getattr__(self, item):
        logging.debug("Pass unknown method request to Hoc")
        return getattr(self._hoc, item)

    def reset_helpers(self):
        ShotNoise.stimCount = 0
        Noise.stimCount = 0
        OrnsteinUhlenbeck.stimCount = 0

    @classmethod
    def register_type(cls, stim_class):
        """ Registers a new class as a handler for a new stim type """
        cls._stim_types[stim_class.__name__] = stim_class
        return stim_class


class BaseStim:
    """
    Barebones stimulus class
    """

    IsPythonOnly = False
    IsNoise = False

    def __init__(self, _target, stim_info: dict, _cell_manager):
        self.duration = float(stim_info["Duration"])  # duration [ms]
        self.delay = float(stim_info["Delay"])        # start time [ms]


@StimulusManager.register_type
class OrnsteinUhlenbeck(BaseStim):
    """
    Ornstein-Uhlenbeck process, injected as current or conductance
    """
    IsPythonOnly = True
    IsNoise = True
    stimCount = 0  # global count for seeding

    def __init__(self, target, stim_info: dict, cell_manager):
        super().__init__(target, stim_info, cell_manager)

        self.stimList = []  # sources go here

        if not self.parse_check_all_parameters(stim_info):
            return None  # nothing to do, stim is a no-op

        # setup random seeds
        seed1 = OrnsteinUhlenbeck.stimCount + 2997  # stimulus block seed
        seed2 = SimConfig.rng_info.getStimulusSeed() + 291204  # stimulus type seed
        seed3 = (lambda x: x + 123) if self.seed is None else (lambda x: self.seed)  # GID seed

        # apply stim to each point in target
        tpoints = target.getPointList(cell_manager)
        for tpoint_list in tpoints:
            gid = tpoint_list.gid
            cell = cell_manager.get_cell(gid)

            self.compute_parameters(cell)

            for sec_id, sc in enumerate(tpoint_list.sclst):
                # skip sections not in this split
                if not sc.exists():
                    continue

                rng = random.Random123(seed1, seed2, seed3(gid))  # setup RNG
                ou_args = (self.tau, self.sigma, self.mean, self.duration)
                ou_kwargs = {'dt': self.dt, 'delay': self.delay, 'rng': rng}
                # inject Ornstein-Uhlenbeck signal
                if stim_info["Mode"] == "Conductance":
                    cs = ConductanceSource.ornstein_uhlenbeck(*ou_args, **ou_kwargs,
                                                              base_amp=self.reversal)
                else:
                    cs = CurrentSource.ornstein_uhlenbeck(*ou_args, **ou_kwargs)
                # attach source to section
                cs.attach_to(sc.sec, tpoint_list.x[sec_id])
                self.stimList.append(cs)  # save source

        OrnsteinUhlenbeck.stimCount += 1  # increment global count

    def parse_check_all_parameters(self, stim_info: dict):
        self.dt = float(stim_info.get("Dt", 0.25))  # stimulus timestep [ms]
        if self.dt <= 0:
            raise Exception("%s time-step must be positive" % self.__class__.__name__)

        self.reversal = float(stim_info.get("Reversal", 0.0))  # reversal potential [mV]

        if stim_info["Mode"] not in ["Current", "Conductance"]:
            raise Exception("%s must be used with mode Current or Conductance"
                            % self.__class__.__name__)

        self.tau = float(stim_info["Tau"])  # relaxation time [ms]
        if self.tau < 0:
            raise Exception("%s relaxation time must be non-negative" % self.__class__.__name__)

        # parse and check stimulus-specific parameters
        if not self.parse_check_stim_parameters(stim_info):
            return False  # nothing to do, stim is a no-op

        self.seed = stim_info.get("Seed")  # random seed override
        if self.seed is not None:
            self.seed = int(self.seed)

        return True

    def parse_check_stim_parameters(self, stim_info):
        self.sigma = float(stim_info["Sigma"])  # signal stdev [uS]
        if self.sigma <= 0:
            raise Exception("%s standard deviation must be positive" % self.__class__.__name__)

        self.mean = float(stim_info["Mean"])    # signal mean [uS]
        if self.mean < 0 and abs(self.mean) > 2 * self.sigma:
            logging.warning("%s signal is mostly zero" % self.__class__.__name__)

        return True

    def compute_parameters(self, cell):
        # nothing to do
        pass


@StimulusManager.register_type
class RelativeOrnsteinUhlenbeck(OrnsteinUhlenbeck):
    """
    Ornstein-Uhlenbeck process, injected as current or conductance,
    relative to cell threshold current or inverse input resistance
    """
    IsPythonOnly = True
    IsNoise = True

    def __init__(self, target, stim_info: dict, cell_manager):
        super().__init__(target, stim_info, cell_manager)

    def parse_check_stim_parameters(self, stim_info):
        self.mean_perc = float(stim_info["MeanPercent"])
        self.sigma_perc = float(stim_info["SDPercent"])

        if stim_info["Mode"] == "Current":
            self.get_relative = lambda x: x.getThreshold()
        else:
            self.get_relative = lambda x: 1.0 / x.input_resistance

        return True

    def compute_parameters(self, cell):
        # threshold current [nA] or inverse input resistance [uS]
        rel_prop = self.get_relative(cell)

        self.sigma = (self.sigma_perc / 100) * rel_prop  # signal stdev [nA or uS]
        if self.sigma <= 0:
            raise Exception("%s standard deviation must be positive" % self.__class__.__name__)

        self.mean = (self.mean_perc / 100) * rel_prop    # signal mean [nA or uS]
        if self.mean < 0 and abs(self.mean) > 2 * self.sigma:
            logging.warning("%s signal is mostly zero" % self.__class__.__name__)

        return True


@StimulusManager.register_type
class ShotNoise(BaseStim):
    """
    ShotNoise stimulus handler implementing Poisson shot noise
    with bi-exponential response and gamma-distributed amplitudes
    """
    IsPythonOnly = True
    IsNoise = True
    stimCount = 0  # global count for seeding

    def __init__(self, target, stim_info: dict, cell_manager):
        super().__init__(target, stim_info, cell_manager)

        self.stimList = []  # CurrentSource's go here

        if not self.parse_check_all_parameters(stim_info):
            return None  # nothing to do, stim is a no-op

        # setup random seeds
        seed1 = ShotNoise.stimCount + 2997  # stimulus block seed
        seed2 = SimConfig.rng_info.getStimulusSeed() + 19216  # stimulus type seed
        seed3 = (lambda x: x + 123) if self.seed is None else (lambda x: self.seed)  # GID seed

        # apply stim to each point in target
        tpoints = target.getPointList(cell_manager)
        for tpoint_list in tpoints:
            gid = tpoint_list.gid
            cell = cell_manager.get_cell(gid)

            self.compute_parameters(cell)

            for sec_id, sc in enumerate(tpoint_list.sclst):
                # skip sections not in this split
                if not sc.exists():
                    continue

                rng = random.Random123(seed1, seed2, seed3(gid))  # setup RNG
                shotnoise_args = (self.tau_D, self.tau_R, self.rate,
                                  self.amp_mean, self.amp_var, self.duration)
                shotnoise_kwargs = {'dt': self.dt, 'delay': self.delay, 'rng': rng}
                # generate shot noise current source
                if stim_info["Mode"] == "Conductance":
                    cs = ConductanceSource.shot_noise(*shotnoise_args, **shotnoise_kwargs,
                                                      base_amp=self.reversal)
                else:
                    cs = CurrentSource.shot_noise(*shotnoise_args, **shotnoise_kwargs)
                # attach current source to section
                cs.attach_to(sc.sec, tpoint_list.x[sec_id])
                self.stimList.append(cs)  # save CurrentSource

        ShotNoise.stimCount += 1  # increment global count

    def parse_check_all_parameters(self, stim_info: dict):
        if stim_info["Mode"] not in ["Current", "Conductance"]:
            raise Exception("%s must be used with mode Current or Conductance"
                            % self.__class__.__name__)

        self.reversal = float(stim_info.get("Reversal", 0.0))  # reversal potential [mV]

        # time parameters
        self.dt = float(stim_info.get("Dt", 0.25))    # stimulus timestep [ms]
        if self.dt <= 0:
            raise Exception("%s time-step must be positive" % self.__class__.__name__)

        ntstep = int(self.duration / self.dt)  # number of timesteps [1]
        if ntstep == 0:
            return False  # nothing to do, stim is a no-op

        # bi-exponential parameters
        self.tau_R = float(stim_info["RiseTime"])     # rise time [ms]
        self.tau_D = float(stim_info["DecayTime"])    # decay time [ms]
        if self.tau_R >= self.tau_D:
            raise Exception("%s bi-exponential rise time must be smaller than decay time"
                            % self.__class__.__name__)

        # parse and check stimulus-specific parameters
        if not self.parse_check_stim_parameters(stim_info):
            return False  # nothing to do, stim is a no-op

        # seed
        self.seed = stim_info.get("Seed")  # random seed override
        if self.seed is not None:
            self.seed = int(self.seed)

        return True

    def parse_check_stim_parameters(self, stim_info: dict):
        """
        Parse parameters for ShotNoise stimulus
        """
        # event rate of Poisson process [Hz]
        self.rate = float(stim_info["Rate"])

        # mean amplitude of shots [nA or uS]
        # when negative we invert the sign of the current
        self.amp_mean = float(stim_info["AmpMean"])
        if self.amp_mean == 0:
            raise Exception("%s amplitude mean must be non-zero" % self.__class__.__name__)

        # variance of amplitude of shots [nA^2 or uS^2]
        self.amp_var = float(stim_info["AmpVar"])
        if self.amp_var <= 0:
            raise Exception("%s amplitude variance must be positive" % self.__class__.__name__)

        return self.rate > 0  # no-op if rate == 0

    def compute_parameters(self, cell):
        # nothing to do
        pass

    def params_from_mean_var(self, mean, var):
        """
        Compute bi-exponential shot noise parameters from desired mean and variance of signal.

        Analytical result derived from a generalization of Campbell's theorem present in
        Rice, S.O., "Mathematical Analysis of Random Noise", BSTJ 23, 3 Jul 1944.
        """
        from math import exp, log

        # bi-exponential time to peak [ms]
        t_peak = log(self.tau_D / self.tau_R) / (1 / self.tau_R - 1 / self.tau_D)
        # bi-exponential peak height [1]
        x_peak = exp(-t_peak / self.tau_D) - exp(-t_peak / self.tau_R)

        rate_ms = (1 + self.cv_square) / 2 * (mean ** 2 / var) / (self.tau_D + self.tau_R)
        self.rate = rate_ms * 1000  # rate in 1 / s [Hz]
        self.amp_mean = mean * x_peak / rate_ms / (self.tau_D - self.tau_R)
        self.amp_var = self.cv_square * self.amp_mean ** 2


@StimulusManager.register_type
class RelativeShotNoise(ShotNoise):
    """
    RelativeShotNoise stimulus handler, same as ShotNoise
    but parameters relative to cell threshold current or inverse input resistance
    """
    IsPythonOnly = True
    IsNoise = True

    def __init__(self, target, stim_info: dict, cell_manager):
        super().__init__(target, stim_info, cell_manager)

    def parse_check_stim_parameters(self, stim_info: dict):
        """
        Parse parameters for RelativeShotNoise stimulus
        """
        # signal mean as percent of cell's threshold [1],
        # when negative we invert the sign of the current
        self.mean_perc = float(stim_info["MeanPercent"])

        # signal standard deviation as percent of cell's threshold [1]
        self.sd_perc = float(stim_info["SDPercent"])
        if self.sd_perc <= 0:
            raise Exception("%s stdev percent must be positive" % self.__class__.__name__)
        if self.sd_perc < 1:
            logging.warning("%s stdev percent too small gives a very high event rate"
                            % self.__class__.__name__)

        # coefficient of variation of shot amplitudes [1]
        cv = float(stim_info["AmpCV"])
        if cv <= 0:
            raise Exception("%s amplitude CV must be positive" % self.__class__.__name__)
        self.cv_square = cv * cv

        if stim_info["Mode"] == "Current":
            self.get_relative = lambda x: x.getThreshold()
        else:
            self.get_relative = lambda x: 1.0 / x.input_resistance

        return self.mean_perc != 0  # no-op if mean_perc == 0

    def compute_parameters(self, cell):
        # threshold current [nA] or inverse input resistance [uS]
        rel_prop = self.get_relative(cell)
        mean = self.mean_perc / 100 * rel_prop  # desired mean [nA or uS]
        sd = self.sd_perc / 100 * rel_prop      # desired standard deviation [nA or uS]
        var = sd * sd                           # variance [nA^2 or uS^2]
        super().params_from_mean_var(mean, var)


@StimulusManager.register_type
class AbsoluteShotNoise(ShotNoise):
    """
    AbsoluteShotNoise stimulus handler, same as ShotNoise
    but parameters from given mean and std. dev.
    """
    IsPythonOnly = True
    IsNoise = True

    def __init__(self, target, stim_info: dict, cell_manager):
        super().__init__(target, stim_info, cell_manager)

    def parse_check_stim_parameters(self, stim_info: dict):
        """
        Parse parameters for AbsoluteShotNoise stimulus
        """
        # signal mean [nA]
        self.mean = float(stim_info["Mean"])

        # signal standard deviation [nA]
        self.sd = float(stim_info["Sigma"])
        if self.sd <= 0:
            raise Exception("%s stdev must be positive" % self.__class__.__name__)

        # coefficient of variation of shot amplitudes [1]
        cv = float(stim_info["AmpCV"])
        if cv <= 0:
            raise Exception("%s amplitude CV must be positive" % self.__class__.__name__)
        self.cv_square = cv * cv

        return True

    def compute_parameters(self, cell):
        super().params_from_mean_var(self.mean, self.sd * self.sd)


@StimulusManager.register_type
class Linear(BaseStim):
    """
    Injects a linear current ramp.
    """
    def __init__(self, target, stim_info: dict, cell_manager):
        super().__init__(target, stim_info, cell_manager)

        self.stimList = []  # CurrentSource's go here

        if not self.parse_check_all_parameters(stim_info):
            return None  # nothing to do, stim is a no-op

        # apply stim to each point in target
        tpoints = target.getPointList(cell_manager)
        for tpoint_list in tpoints:
            gid = tpoint_list.gid
            cell = cell_manager.get_cell(gid)

            self.compute_parameters(cell)

            for sec_id, sc in enumerate(tpoint_list.sclst):
                # skip sections not in this split
                if not sc.exists():
                    continue

                # generate ramp current source
                cs = CurrentSource.ramp(self.amp_start, self.amp_end, self.duration,
                                        delay=self.delay)
                # attach current source to section
                cs.attach_to(sc.sec, tpoint_list.x[sec_id])
                self.stimList.append(cs)  # save CurrentSource

    def parse_check_all_parameters(self, stim_info: dict):
        # Amplitude at start
        self.amp_start = float(stim_info["AmpStart"])

        # Amplitude at end (optional, else same as start)
        self.amp_end = float(stim_info.get("AmpEnd", self.amp_start))

        return self.amp_start != 0 or self.amp_end != 0  # no-op if both 0

    def compute_parameters(self, cell):
        pass  # nothing to do


@StimulusManager.register_type
class Hyperpolarizing(Linear):
    """
    Injects a constant step with a cell's hyperpolarizing current.
    """
    def __init__(self, target, stim_info: dict, cell_manager):
        super().__init__(target, stim_info, cell_manager)

    def parse_check_all_parameters(self, stim_info: dict):
        return True

    def compute_parameters(self, cell):
        hypamp = cell.getHypAmp()
        self.amp_start = hypamp
        self.amp_end = hypamp


@StimulusManager.register_type
class RelativeLinear(Linear):
    """
    Injects a linear current ramp relative to cell threshold.
    """
    def __init__(self, target, stim_info: dict, cell_manager):
        super().__init__(target, stim_info, cell_manager)

    def parse_check_all_parameters(self, stim_info: dict):
        # Amplitude at start as percent of threshold
        self.perc_start = float(stim_info["PercentStart"])

        # Amplitude at end as percent of threshold (optional, else same as start)
        self.perc_end = float(stim_info.get("PercentEnd", self.perc_start))

        return self.perc_start != 0 or self.perc_end != 0  # no-op if both 0

    def compute_parameters(self, cell):
        threshold = cell.getThreshold()
        # here we use parentheses to match HOC exactly
        self.amp_start = threshold * (self.perc_start / 100)
        self.amp_end = threshold * (self.perc_end / 100)


@StimulusManager.register_type
class SubThreshold(Linear):
    """
    Injects a current step at some percent below a cell's threshold.
    """
    def __init__(self, target, stim_info: dict, cell_manager):
        super().__init__(target, stim_info, cell_manager)

    def parse_check_all_parameters(self, stim_info: dict):
        # amplitude as percent below threshold = 100%
        self.perc_less = float(stim_info["PercentLess"])

        return True

    def compute_parameters(self, cell):
        threshold = cell.getThreshold()
        # here we use parentheses to match HOC exactly
        self.amp_start = threshold * (100 - self.perc_less) / 100
        self.amp_end = self.amp_start


@StimulusManager.register_type
class Noise(BaseStim):
    """
    Inject a noisy (gaussian) current step, relative to cell threshold or not.
    """
    IsNoise = True
    stimCount = 0  # global count for seeding

    def __init__(self, target, stim_info: dict, cell_manager):
        super().__init__(target, stim_info, cell_manager)

        self.stimList = []  # CurrentSource's go here

        self.parse_check_all_parameters(stim_info)

        sim_dt = float(SimConfig.run_conf["Dt"])  # simulation time-step [ms]
        rng_mode = SimConfig.rng_info.getRNGMode()  # simulation RNGMode

        # setup RNG
        if rng_mode == SimConfig.rng_info.COMPATIBILITY:
            rand = lambda gid: random.RNG(seed=gid + Noise.stimCount)
        elif rng_mode == SimConfig.rng_info.UPMCELLRAN4:
            rand = lambda gid: random.MCellRan4(Noise.stimCount * 10000 + 100,
                                                SimConfig.rng_info.getGlobalSeed() +
                                                SimConfig.rng_info.getStimulusSeed() +
                                                gid * 1000)
        elif rng_mode == SimConfig.rng_info.RANDOM123:
            rand = lambda gid: random.Random123(Noise.stimCount + 100,
                                                SimConfig.rng_info.getStimulusSeed() + 500,
                                                gid + 300)

        # apply stim to each point in target
        tpoints = target.getPointList(cell_manager)
        for tpoint_list in tpoints:
            gid = tpoint_list.gid
            cell = cell_manager.get_cell(gid)

            self.compute_parameters(cell)

            rng = rand(gid)  # setup RNG
            # draw already used numbers
            if rng_mode != SimConfig.rng_info.COMPATIBILITY and self.delay > 0:
                self.draw_already_used_numbers(rng, sim_dt)

            for sec_id, sc in enumerate(tpoint_list.sclst):
                # skip sections not in this split
                if not sc.exists():
                    continue

                # generate noise current source
                cs = CurrentSource.noise(self.mean, self.var, self.duration,
                                         dt=self.dt, delay=self.delay, rng=rng)
                # attach current source to section
                cs.attach_to(sc.sec, tpoint_list.x[sec_id])
                self.stimList.append(cs)  # save CurrentSource

        Noise.stimCount += 1  # increment global count

    def parse_check_all_parameters(self, stim_info: dict):
        self.dt = float(stim_info.get("Dt", 0.5))  # stimulus timestep [ms]
        if self.dt <= 0:
            raise Exception("%s time-step must be positive" % self.__class__.__name__)

        if "Mean" in stim_info:
            self.is_relative = False
            self.mean = float(stim_info["Mean"])  # noise current mean [nA]

            self.var = float(stim_info["Variance"])  # noise current variance [nA]
            if self.var <= 0:
                raise Exception("%s variance must be positive" % self.__class__.__name__)
        else:
            self.is_relative = True
            # noise current mean as percent of threshold
            self.mean_perc = float(stim_info["MeanPercent"])

            # noise current variance as percent of threshold
            self.var_perc = float(stim_info["Variance"])
            if self.var_perc <= 0:
                raise Exception("%s variance percent must be positive" % self.__class__.__name__)

        return True

    def compute_parameters(self, cell):
        if self.is_relative:
            threshold = cell.getThreshold()  # threshold current [nA]
            # here threshold MUST be first factor to match HOC exactly
            self.mean = threshold * self.mean_perc / 100
            # note that here variance has units of nA, not nA^2
            self.var = threshold * self.var_perc / 100

    def draw_already_used_numbers(self, rng, dt):
        prev_t = 0
        tstep = self.duration - dt

        while prev_t < self.delay - dt:
            if prev_t + tstep < self.delay - dt:
                next_t = prev_t + tstep
            else:
                next_t = self.delay - dt

            tvec = Nd.h.Vector()
            tvec.indgen(prev_t, next_t, self.dt)
            stim = Nd.h.Vector(len(tvec))
            stim.setrand(rng)

            prev_t = next_t + dt


@StimulusManager.register_type
class Pulse(BaseStim):
    """
    Inject a pulse train with given amplitude, frequency and width.
    """
    def __init__(self, target, stim_info: dict, cell_manager):
        super().__init__(target, stim_info, cell_manager)

        self.stimList = []  # CurrentSource's go here

        if not self.parse_check_all_parameters(stim_info):
            return None  # nothing to do, stim is a no-op

        # apply stim to each point in target
        tpoints = target.getPointList(cell_manager)
        for tpoint_list in tpoints:
            for sec_id, sc in enumerate(tpoint_list.sclst):
                # skip sections not in this split
                if not sc.exists():
                    continue

                # generate pulse train current source
                cs = CurrentSource.train(self.amp, self.freq, self.width,
                                         self.duration, delay=self.delay)
                # attach current source to section
                cs.attach_to(sc.sec, tpoint_list.x[sec_id])
                self.stimList.append(cs)  # save CurrentSource

    def parse_check_all_parameters(self, stim_info: dict):
        self.amp = float(stim_info["AmpStart"])  # amplitude [nA]
        self.freq = float(stim_info["Frequency"])  # frequency [Hz]
        self.width = float(stim_info["Width"])  # pulse width [ms]

        return self.freq > 0 and self.width > 0  # no-op if any is 0


@StimulusManager.register_type
class Sinusoidal(BaseStim):
    """
    Inject a sinusoidal current with given amplitude and frequency.
    """
    def __init__(self, target, stim_info: dict, cell_manager):
        super().__init__(target, stim_info, cell_manager)

        self.stimList = []  # CurrentSource's go here

        if not self.parse_check_all_parameters(stim_info):
            return None  # nothing to do, stim is a no-op

        # apply stim to each point in target
        tpoints = target.getPointList(cell_manager)
        for tpoint_list in tpoints:
            for sec_id, sc in enumerate(tpoint_list.sclst):
                # skip sections not in this split
                if not sc.exists():
                    continue

                # generate sinusoidal current source
                cs = CurrentSource.sin(self.amp, self.duration, self.freq,
                                       step=self.dt, delay=self.delay)
                # attach current source to section
                cs.attach_to(sc.sec, tpoint_list.x[sec_id])
                self.stimList.append(cs)  # save CurrentSource

    def parse_check_all_parameters(self, stim_info: dict):
        self.dt = float(stim_info.get("Dt", 0.025))  # stimulus timestep [ms]
        if self.dt <= 0:
            raise Exception("%s time-step must be positive" % self.__class__.__name__)

        self.amp = float(stim_info["AmpStart"])  # amplitude [nA]
        self.freq = float(stim_info["Frequency"])  # frequency [Hz]

        return self.freq > 0  # no-op if 0


@StimulusManager.register_type
class SEClamp(BaseStim):
    """
    Apply a single electrode voltage clamp.
    """
    def __init__(self, target, stim_info: dict, cell_manager):
        super().__init__(target, stim_info, cell_manager)

        self.stimList = []  # SEClamp's go here

        self.parse_check_all_parameters(stim_info)

        # apply stim to each point in target
        tpoints = target.getPointList(cell_manager)
        for tpoint_list in tpoints:
            for sec_id, sc in enumerate(tpoint_list.sclst):
                # skip sections not in this split
                if not sc.exists():
                    continue

                # create single electrode voltage clamp at location
                seclamp = Nd.h.SEClamp(tpoint_list.x[sec_id], sec=sc.sec)
                seclamp.rs = self.rs
                seclamp.dur1 = self.duration
                seclamp.amp1 = self.vhold
                self.stimList.append(seclamp)  # save SEClamp

    def parse_check_all_parameters(self, stim_info: dict):
        self.vhold = float(stim_info["Voltage"])  # holding voltage [mV]
        self.rs = float(stim_info.get("RS", 0.01))  # series resistance [MOhm]
        if self.delay > 0:
            logging.warning("%s ignores delay" % self.__class__.__name__)
